from datetime import timedelta

import isodate
from flask_json import as_json
from flask import request, current_app

from flexmeasures.utils.entity_address_utils import (
    parse_entity_address,
    EntityAddressException,
)
from flexmeasures.api.common.responses import (
    invalid_domain,
    invalid_datetime,
    invalid_timezone,
    request_processed,
    unrecognized_event,
    invalid_market,
    unknown_prices,
    unrecognized_connection_group,
    outdated_event_id,
    ptus_incomplete,
)
from flexmeasures.api.common.utils.api_utils import (
    groups_to_dict,
    get_form_from_request,
)
from flexmeasures.api.common.utils.validators import (
    type_accepted,
    assets_required,
    optional_duration_accepted,
    usef_roles_accepted,
    units_accepted,
    parse_isodate_str,
)
from flexmeasures.data.config import db
from flexmeasures.data.models.assets import Asset
from flexmeasures.data.models.planning.battery import schedule_battery
from flexmeasures.data.models.planning.exceptions import UnknownPricesException
from flexmeasures.data.services.resources import has_assets, can_access_asset


@type_accepted("GetDeviceMessageRequest")
@assets_required("event")
@optional_duration_accepted(timedelta(hours=6))
@as_json
def get_device_message_response(generic_asset_name_groups, duration):

    unit = "MW"
    min_planning_horizon = timedelta(
        hours=24
    )  # user can request a shorter planning, but the scheduler takes into account at least this horizon
    planning_horizon = min(
        max(min_planning_horizon, duration),
        current_app.config.get("FLEXMEASURES_PLANNING_HORIZON"),
    )

    if not has_assets():
        current_app.logger.info("User doesn't seem to have any assets.")

    value_groups = []
    new_event_groups = []
    for event_group in generic_asset_name_groups:
        for event in event_group:

            # Parse the entity address
            try:
                ea = parse_entity_address(event, entity_type="event", fm_scheme="fm0")
            except EntityAddressException as eae:
                return invalid_domain(str(eae))
            asset_id = ea["asset_id"]
            event_id = ea["event_id"]
            event_type = ea["event_type"]

            # Look for the Asset object
            asset = Asset.query.filter(Asset.id == asset_id).one_or_none()
            if asset is None or not can_access_asset(asset):
                current_app.logger.warning(
                    "Cannot identify asset %s given the event." % event
                )
                return unrecognized_connection_group()
            if asset.asset_type_name != "battery":
                return invalid_domain(
                    "API version 1.2 only supports device messages for batteries. Asset ID:%s is not a battery."
                    % asset_id
                )
            if event_type != "soc" or event_id != asset.soc_udi_event_id:
                return unrecognized_event(event_id, event_type)
            start = asset.soc_datetime
            resolution = asset.event_resolution

            # Look for the Market object
            market = asset.market
            if market is None:
                return invalid_market()

            # Schedule the asset
            try:
                schedule = schedule_battery(
                    asset,
                    market,
                    start,
                    start + planning_horizon,
                    resolution,
                    soc_at_start=asset.soc_in_mwh,
                    prefer_charging_sooner=False,
                )
            except UnknownPricesException:
                return unknown_prices()
            else:
                # Update the planning window
                start = schedule.index[0]
                duration = min(duration, schedule.index[-1] + resolution - start)
                schedule = schedule[start : start + duration - resolution]
            value_groups.append(schedule.tolist())
            new_event_groups.append(event)

    response = groups_to_dict(
        new_event_groups, value_groups, generic_asset_type_name="event"
    )
    response["start"] = isodate.datetime_isoformat(start)
    response["duration"] = isodate.duration_isoformat(duration)
    response["unit"] = unit

    d, s = request_processed()
    return dict(**response, **d), s


@usef_roles_accepted("Prosumer")
@type_accepted("PostUdiEventRequest")
@units_accepted("State of charge", "kWh", "MWh")
@as_json
def post_udi_event_response(unit):  # noqa: C901

    if not has_assets():
        current_app.logger.info("User doesn't seem to have any assets.")

    form = get_form_from_request(request)

    # check datetime, or use server_now
    if "datetime" not in form:
        return invalid_datetime("Missing datetime parameter.")
    else:
        datetime = parse_isodate_str(form.get("datetime"))
        if datetime is None:
            return invalid_datetime(
                "Cannot parse datetime string %s as iso date" % form.get("datetime")
            )
        if datetime.tzinfo is None:
            current_app.logger.warning(
                "Cannot parse timezone of 'datetime' value %s" % form.get("datetime")
            )
            return invalid_timezone("Datetime should explicitly state a timezone.")

    # parse event/address info
    if "event" not in form:
        return invalid_domain("No event identifier sent.")
    try:
        ea = parse_entity_address(
            form.get("event"), entity_type="event", fm_scheme="fm0"
        )
    except EntityAddressException as eae:
        return invalid_domain(str(eae))

    asset_id = ea["asset_id"]
    event_id = ea["event_id"]
    event_type = ea["event_type"]

    if event_type != "soc":
        return unrecognized_event(event_id, event_type)

    # get asset
    asset: Asset = Asset.query.filter_by(id=asset_id).one_or_none()
    if asset is None or not can_access_asset(asset):
        current_app.logger.warning("Cannot identify asset via %s." % ea)
        return unrecognized_connection_group()
    if asset.asset_type_name != "battery":
        return invalid_domain(
            "API version 1.2 only supports UDI events for batteries. Asset ID:%s is not a battery."
            % asset_id
        )

    # unless on play, keep events ordered by entry date and ID
    if current_app.config.get("FLEXMEASURES_MODE") != "play":
        # do not allow new date to be after last date
        if asset.soc_datetime is not None:
            if asset.soc_datetime >= datetime:
                msg = (
                    "The date of the requested UDI event (%s) is earlier than the latest known date (%s)."
                    % (datetime, asset.soc_datetime)
                )
                current_app.logger.warning(msg)
                return invalid_datetime(msg)

        # check if udi event id is higher than existing
        if asset.soc_udi_event_id is not None:
            if asset.soc_udi_event_id >= event_id:
                return outdated_event_id(event_id, asset.soc_udi_event_id)

    # get value
    if "value" not in form:
        return ptus_incomplete()
    value = form.get("value")
    if unit == "kWh":
        value = value / 1000.0

    # store new soc in asset
    asset.soc_datetime = datetime
    asset.soc_udi_event_id = event_id
    asset.soc_in_mwh = value

    db.session.commit()
    return request_processed("Request has been processed.")



"""Useful test messages"""
from datetime import timedelta
from typing import List, Optional, Union

from isodate import duration_isoformat, parse_datetime, parse_duration
from numpy import tile
import pandas as pd

from flexmeasures.api.common.utils.validators import validate_user_sources
from flexmeasures.data.models.assets import Power, Asset


def message_for_get_meter_data(
    no_connection: bool = False,
    invalid_connection: bool = False,
    single_connection: bool = False,
    demo_connection: bool = False,
    invalid_unit: bool = False,
    no_unit: bool = False,
    resolution: str = "",
    source: Optional[Union[str, List[str]]] = None,
) -> dict:
    message = {
        "type": "GetMeterDataRequest",
        "start": "2015-01-01T00:00:00Z",
        "duration": "PT1H30M",
        "connections": ["CS 1", "CS 2", "CS 3"],
        "unit": "MW",
    }
    if no_connection:
        message.pop("connections", None)
    elif invalid_connection:
        message["connections"] = ["Non-existing asset 1", "Non-existing asset 2"]
    elif single_connection:
        message["connection"] = message["connections"][0]
        message.pop("connections", None)
    elif demo_connection:
        message["connection"] = "CS 0"
        message.pop("connections", None)
    if no_unit:
        message.pop("unit", None)
    elif invalid_unit:
        message["unit"] = "MW/h"
    if resolution:
        message["resolution"] = resolution
    if source:
        message["source"] = source
    return message


def message_for_post_meter_data(
    no_connection: bool = False,
    single_connection: bool = False,
    single_connection_group: bool = False,
    production: bool = False,
    different_target_resolutions: bool = False,
    tile_n=1,
) -> dict:
    sign = 1 if production is False else -1
    message = {
        "type": "PostMeterDataRequest",
        "groups": [
            {
                "connections": ["CS 1", "CS 2"],
                "values": (
                    tile([306.66, 306.66, 0, 0, 306.66, 306.66], tile_n) * sign
                ).tolist(),
            },
            {
                "connection": ["CS 4" if different_target_resolutions else "CS 3"],
                "values": (
                    tile([306.66, 0, 0, 0, 306.66, 306.66], tile_n) * sign
                ).tolist(),
            },
        ],
        "start": "2015-01-01T00:00:00Z",
        "duration": duration_isoformat(timedelta(hours=1.5 * tile_n)),
        "horizon": "PT0H",
        "unit": "MW",
    }
    if no_connection:
        message.pop("groups", None)
    elif single_connection:
        message["connection"] = message["groups"][0]["connections"][0]
        message["values"] = message["groups"][1]["values"]
        message.pop("groups", None)
    elif single_connection_group:
        message["connections"] = message["groups"][0]["connections"]
        message["values"] = message["groups"][0]["values"]
        message.pop("groups", None)

    return message


def count_connections_in_post_message(message: dict) -> int:
    connections = 0
    if "groups" in message:
        message = dict(
            connections=message["groups"][0]["connections"],
            connection=message["groups"][1]["connection"],
        )
    if "connection" in message:
        connections += 1
    if "connections" in message:
        connections += len(message["connections"])
    return connections


def verify_power_in_db(
    message, asset, expected_df: pd.DataFrame, db, swapped_sign: bool = False
):
    """util method to verify that power data ended up in the database"""
    # todo: combine with verify_prices_in_db (in v1_1 utils) into a single function (NB different horizon filters)
    start = parse_datetime(message["start"])
    end = start + parse_duration(message["duration"])
    horizon = (
        parse_duration(message["horizon"]) if "horizon" in message else timedelta(0)
    )
    resolution = asset.event_resolution
    query = (
        db.session.query(Power.datetime, Power.value, Power.data_source_id)
        .filter((Power.datetime > start - resolution) & (Power.datetime < end))
        .filter(Power.horizon == horizon)
        .join(Asset)
        .filter(Asset.name == asset.name)
    )
    if "source" in message:
        source_ids = validate_user_sources(message["source"])
        query = query.filter(Power.data_source_id.in_(source_ids))
    df = pd.DataFrame(
        query.all(), columns=[col["name"] for col in query.column_descriptions]
    )
    df = df.set_index(["datetime", "data_source_id"]).sort_index()
    if swapped_sign:
        df["value"] = -df["value"]

    assert df["value"].to_list() == expected_df["value"].to_list()



from datetime import timedelta

import pytest
from flask import url_for
from iso8601 import parse_date
from numpy import repeat

from flexmeasures.api.common.utils.api_utils import message_replace_name_with_ea
from flexmeasures.api.tests.utils import get_auth_token
from flexmeasures.api.v1.tests.utils import (
    message_for_post_meter_data,
    message_for_get_meter_data,
    count_connections_in_post_message,
)
from flexmeasures.data.models.assets import Asset


@pytest.mark.parametrize(
    "post_message",
    [
        message_for_post_meter_data(),
        message_for_post_meter_data(single_connection=True),
        message_for_post_meter_data(single_connection_group=True),
    ],
)
@pytest.mark.parametrize(
    "get_message",
    [
        message_for_get_meter_data(),
        message_for_get_meter_data(single_connection=False),
        message_for_get_meter_data(resolution="PT30M"),
    ],
)
def test_post_and_get_meter_data(
    setup_fresh_api_test_data, app, clean_redis, client, post_message, get_message
):
    """
    Tries to post meter data as a logged-in test user with the MDC role, which should succeed.
    There should be some ForecastingJobs waiting now.
    Then tries to get meter data, which should succeed, and should return the same meter data as was posted,
    or a downsampled version, if that was requested.
    """

    # post meter data
    auth_token = get_auth_token(client, "test_prosumer@seita.nl", "testtest")
    post_meter_data_response = client.post(
        url_for("flexmeasures_api_v1.post_meter_data"),
        json=message_replace_name_with_ea(post_message),
        headers={"Authorization": auth_token},
    )
    print("Server responded with:\n%s" % post_meter_data_response.json)
    assert post_meter_data_response.status_code == 200
    assert post_meter_data_response.json["type"] == "PostMeterDataResponse"

    # look for Forecasting jobs
    expected_connections = count_connections_in_post_message(post_message)
    assert (
        len(app.queues["forecasting"]) == 4 * expected_connections
    )  # four horizons times the number of assets
    horizons = repeat(
        [
            timedelta(hours=1),
            timedelta(hours=6),
            timedelta(hours=24),
            timedelta(hours=48),
        ],
        expected_connections,
    )
    jobs = sorted(app.queues["forecasting"].jobs, key=lambda x: x.kwargs["horizon"])
    for job, horizon in zip(jobs, horizons):
        assert job.kwargs["horizon"] == horizon
        assert job.kwargs["start"] == parse_date(post_message["start"]) + horizon
    for asset_name in ("CS 1", "CS 2", "CS 3"):
        if asset_name in str(post_message):
            asset = Asset.query.filter_by(name=asset_name).one_or_none()
            assert asset.id in [job.kwargs["asset_id"] for job in jobs]

    # get meter data
    get_meter_data_response = client.get(
        url_for("flexmeasures_api_v1.get_meter_data"),
        query_string=message_replace_name_with_ea(get_message),
        headers={"Authorization": auth_token},
    )
    print("Server responded with:\n%s" % get_meter_data_response.json)
    assert get_meter_data_response.status_code == 200
    assert get_meter_data_response.json["type"] == "GetMeterDataResponse"
    if "groups" in post_message:
        posted_values = post_message["groups"][0]["values"]
    else:
        posted_values = post_message["values"]
    if "groups" in get_meter_data_response.json:
        gotten_values = get_meter_data_response.json["groups"][0]["values"]
    else:
        gotten_values = get_meter_data_response.json["values"]

    if "resolution" not in get_message or get_message["resolution"] == "":
        assert gotten_values == posted_values
    else:
        # We used a target resolution of 30 minutes, so double of 15 minutes.
        # Six values went in, three come out.
        if posted_values[1] > 0:  # see utils.py:message_for_post_meter_data
            assert gotten_values == [306.66, -0.0, 306.66]
        else:
            assert gotten_values == [153.33, 0, 306.66]



from importlib_metadata import version, PackageNotFoundError


__version__ = "Unknown"

# This uses importlib.metadata behaviour added in Python 3.8
# and relies on setuptools_scm.
try:
    __version__ = version("flexmeasures")
except PackageNotFoundError:
    # package is not installed
    pass



from inspect import getmembers, isfunction

from . import belief_charts
from .defaults import apply_chart_defaults


def chart_type_to_chart_specs(chart_type: str, **kwargs) -> dict:
    """Create chart specs of a given chart type, using FlexMeasures defaults for settings like width and height.

    :param chart_type:  Name of a variable defining chart specs or a function returning chart specs.
                        The chart specs can be a dictionary or an Altair chart specification.
                        - In case of a dictionary, the creator needs to ensure that the dictionary contains valid specs
                        - In case of an Altair chart specification, Altair validates for you
    :returns:           A dictionary containing a vega-lite chart specification
    """
    # Create a dictionary mapping chart types to chart specs, and apply defaults to the chart specs, too.
    belief_charts_mapping = {
        chart_type: apply_chart_defaults(chart_specs)
        for chart_type, chart_specs in getmembers(belief_charts)
        if isfunction(chart_specs) or isinstance(chart_specs, dict)
    }
    # Create chart specs
    chart_specs_or_fnc = belief_charts_mapping[chart_type]
    if isfunction(chart_specs_or_fnc):
        return chart_specs_or_fnc(**kwargs)
    return chart_specs_or_fnc



from webargs.flaskparser import use_args

from flexmeasures.api.common.schemas.sensor_data import SensorDataSchema
from flexmeasures.api.common.utils.api_utils import save_to_db


@use_args(
    SensorDataSchema(),
    location="json",
)
def post_data(sensor_data):
    """POST to /sensorData

    Experimental dev feature which uses timely-beliefs
    to create and save the data structure.
    """
    beliefs = SensorDataSchema.load_bdf(sensor_data)
    response, code = save_to_db(beliefs)
    response.update(type="PostSensorDataResponse")
    return response, code


def get_data():
    """ GET from /sensorData"""
    # - use data.models.time_series.Sensor::search_beliefs() - might need to add a belief_horizon parameter
    # - create the serialize method on the schema, to turn the resulting BeliefsDataFrame
    #   to the JSON the API should respond with.
    pass



from datetime import datetime, timedelta
from typing import Dict, List

from flask import request, session, current_app
from flask_security import roles_accepted
from flask_security.core import current_user
import pandas as pd
import numpy as np
from bokeh.embed import components
import bokeh.palettes as palettes
from bokeh.plotting import Figure

from flexmeasures.data.models.assets import Asset, Power
from flexmeasures.data.models.markets import Price
from flexmeasures.data.queries.portfolio import (
    get_structure,
    get_power_data,
    get_price_data,
)
from flexmeasures.data.services.resources import get_assets
from flexmeasures.utils import time_utils
from flexmeasures.utils.flexmeasures_inflection import capitalize
import flexmeasures.ui.utils.plotting_utils as plotting
from flexmeasures.ui.views import flexmeasures_ui
from flexmeasures.ui.utils.view_utils import (
    render_flexmeasures_template,
    set_time_range_for_session,
)


@flexmeasures_ui.route("/portfolio", methods=["GET", "POST"])
@roles_accepted("admin", "Prosumer")
def portfolio_view():  # noqa: C901
    """Portfolio view.
    By default, this page shows live results (production, consumption and market data) from the user's portfolio.
    Time windows for which the platform has identified upcoming balancing opportunities are highlighted.
    The page can also be used to navigate historical results.
    """

    set_time_range_for_session()
    start = session.get("start_time")
    end = session.get("end_time")
    resolution = session.get("resolution")

    # Get plot perspective
    perspectives = ["production", "consumption"]
    default_stack_side = "production"  # todo: move to user config setting
    show_stacked = request.values.get("show_stacked", default_stack_side)
    perspectives.remove(show_stacked)
    show_summed: str = perspectives[0]
    plot_label = f"Stacked {show_stacked} vs aggregated {show_summed}"

    # Get structure and data
    assets: List[Asset] = get_assets(
        order_by_asset_attribute="display_name", order_direction="asc"
    )
    represented_asset_types, markets, resource_dict = get_structure(assets)
    for resource_name, resource in resource_dict.items():
        resource.load_sensor_data(
            [Power, Price],
            start=start,
            end=end,
            resolution=resolution,
            exclude_source_types=["scheduling script"],
        )  # The resource caches the results
    (
        supply_resources_df_dict,
        demand_resources_df_dict,
        production_per_asset_type,
        consumption_per_asset_type,
        production_per_asset,
        consumption_per_asset,
    ) = get_power_data(resource_dict)
    price_bdf_dict, average_price_dict = get_price_data(resource_dict)

    # Pick a perspective for summing and for stacking
    sum_dict = (
        demand_resources_df_dict.values()
        if show_summed == "consumption"
        else supply_resources_df_dict.values()
    )
    power_sum_df = (
        pd.concat(sum_dict, axis=1).sum(axis=1).to_frame(name="event_value")
        if sum_dict
        else pd.DataFrame()
    )

    # Create summed plot
    power_sum_df = data_or_zeroes(power_sum_df, start, end, resolution)
    x_range = plotting.make_range(
        pd.date_range(start, end, freq=resolution, closed="left")
    )
    fig_profile = plotting.create_graph(
        power_sum_df,
        unit="MW",
        title=plot_label,
        x_range=x_range,
        x_label="Time (resolution of %s)"
        % time_utils.freq_label_to_human_readable_label(resolution),
        y_label="Power (in MW)",
        legend_location="top_right",
        legend_labels=(capitalize(show_summed), None, None),
        show_y_floats=True,
        non_negative_only=True,
    )
    fig_profile.plot_height = 450
    fig_profile.plot_width = 900

    # Create stacked plot
    stack_dict = (
        rename_event_value_column_to_resource_name(supply_resources_df_dict).values()
        if show_summed == "consumption"
        else rename_event_value_column_to_resource_name(
            demand_resources_df_dict
        ).values()
    )
    df_stacked_data = pd.concat(stack_dict, axis=1) if stack_dict else pd.DataFrame()
    df_stacked_data = data_or_zeroes(df_stacked_data, start, end, resolution)
    df_stacked_areas = stack_df(df_stacked_data)

    num_areas = df_stacked_areas.shape[1]
    if num_areas <= 2:
        colors = ["#99d594", "#dddd9d"]
    else:
        colors = palettes.brewer["Spectral"][num_areas]

    df_stacked_data = time_utils.tz_index_naively(df_stacked_data)
    x_points = np.hstack((df_stacked_data.index[::-1], df_stacked_data.index))

    fig_profile.grid.minor_grid_line_color = "#eeeeee"

    for a, area in enumerate(df_stacked_areas):
        fig_profile.patch(
            x_points,
            df_stacked_areas[area].values,
            color=colors[a],
            alpha=0.8,
            line_color=None,
            legend=df_stacked_data.columns[a],
            level="underlay",
        )

    portfolio_plots_script, portfolio_plots_divs = components(fig_profile)

    # Flexibility numbers and a mocked control action are mocked for demo mode at the moment
    flex_info = {}
    if current_app.config.get("FLEXMEASURES_MODE") == "demo":
        flex_info = mock_flex_info(assets, represented_asset_types)
        fig_actions = mock_flex_figure(
            x_range, power_sum_df.index, fig_profile.plot_width
        )
        mock_flex_action_in_main_figure(fig_profile)
        portfolio_plots_script, portfolio_plots_divs = components(
            (fig_profile, fig_actions)
        )

    return render_flexmeasures_template(
        "views/portfolio.html",
        assets=assets,
        average_prices=average_price_dict,
        asset_types=represented_asset_types,
        markets=markets,
        production_per_asset=production_per_asset,
        consumption_per_asset=consumption_per_asset,
        production_per_asset_type=production_per_asset_type,
        consumption_per_asset_type=consumption_per_asset_type,
        sum_production=sum(production_per_asset_type.values()),
        sum_consumption=sum(consumption_per_asset_type.values()),
        flex_info=flex_info,
        portfolio_plots_script=portfolio_plots_script,
        portfolio_plots_divs=portfolio_plots_divs,
        alt_stacking=show_summed,
        fm_mode=current_app.config.get("FLEXMEASURES_MODE"),
    )


def data_or_zeroes(df: pd.DataFrame, start, end, resolution) -> pd.DataFrame:
    """Making really sure we have the structure to let the plots not fail"""
    if df is None or df.empty:
        return pd.DataFrame(
            index=pd.date_range(
                start=start,
                end=end,
                freq=resolution,
                tz=time_utils.get_timezone(),
                closed="left",
            ),
            columns=["event_value"],
        ).fillna(0)
    else:
        return df.fillna(0)


def stack_df(df: pd.DataFrame) -> pd.DataFrame:
    """Stack columns of df cumulatively, include bottom"""
    df_top = df.cumsum(axis=1)
    df_bottom = df_top.shift(axis=1).fillna(0)[::-1]
    df_stack = pd.concat([df_bottom, df_top], ignore_index=True)
    return df_stack


def rename_event_value_column_to_resource_name(
    df_dict: Dict[str, pd.DataFrame]
) -> Dict[str, pd.DataFrame]:
    """Replace the column name "event_source" with the resource name, for each resource in the dictionary."""
    return {
        df_name: df.rename(columns={"event_value": capitalize(df_name)})
        for df_name, df in df_dict.items()
    }


def mock_flex_info(assets, represented_asset_types) -> dict:
    flex_info = dict(
        curtailment_per_asset={a.name: 0.0 for a in assets},
        shifting_per_asset={a.name: 0.0 for a in assets},
        profit_loss_flexibility_per_asset={a.name: 0.0 for a in assets},
        curtailment_per_asset_type={k: 0.0 for k in represented_asset_types.keys()},
        shifting_per_asset_type={k: 0.0 for k in represented_asset_types.keys()},
        profit_loss_flexibility_per_asset_type={
            k: 0.0 for k in represented_asset_types.keys()
        },
    )

    flex_info["shifting_per_asset"]["48_r"] = 1.1
    flex_info["profit_loss_flexibility_per_asset"]["48_r"] = 76000.0
    flex_info["shifting_per_asset_type"]["one-way EVSE"] = flex_info[
        "shifting_per_asset"
    ]["48_r"]
    flex_info["profit_loss_flexibility_per_asset_type"]["one-way EVSE"] = flex_info[
        "profit_loss_flexibility_per_asset"
    ]["48_r"]
    flex_info["curtailment_per_asset"]["hw-onshore"] = 1.3
    flex_info["profit_loss_flexibility_per_asset"]["hw-onshore"] = 84000.0
    flex_info["curtailment_per_asset_type"]["wind turbines"] = flex_info[
        "curtailment_per_asset"
    ]["hw-onshore"]
    flex_info["profit_loss_flexibility_per_asset_type"]["wind turbines"] = flex_info[
        "profit_loss_flexibility_per_asset"
    ]["hw-onshore"]

    flex_info["sum_curtailment"] = sum(flex_info["curtailment_per_asset_type"].values())  # type: ignore
    flex_info["sum_shifting"] = sum(flex_info["shifting_per_asset_type"].values())  # type: ignore
    flex_info["sum_profit_loss_flexibility"] = sum(  # type: ignore
        flex_info["profit_loss_flexibility_per_asset_type"].values()  # type: ignore
    )
    return flex_info


def mock_flex_figure(x_range, x_index, fig_width) -> Figure:
    df_actions = pd.DataFrame(index=x_index, columns=["event_value"]).fillna(0)
    next_action_hour4 = get_flex_action_hour(4)
    if next_action_hour4 in df_actions.index:
        if current_user.is_authenticated:
            if current_user.has_role("admin"):
                df_actions.loc[next_action_hour4] = -2.4  # mock two actions
            elif "wind" in current_user.email:
                df_actions.loc[next_action_hour4] = -1.3  # mock one action
            elif "charging" in current_user.email:
                df_actions.loc[next_action_hour4] = -1.1  # mock one action

    next_action_hour2 = get_flex_action_hour(2)
    if next_action_hour2 in df_actions.index:
        if next_action_hour2 < next_action_hour4 and (
            current_user.is_authenticated
            and (
                current_user.has_role("admin")
                or "wind" in current_user.email
                or "charging" in current_user.email
            )
        ):
            # mock the shift "payback" (actually occurs earlier in our mock example)
            df_actions.loc[next_action_hour2] = 1.1

    next_action_hour9 = get_flex_action_hour(9)
    if next_action_hour9 in df_actions.index:
        # mock some other ordered actions that are not in an opportunity hour anymore
        df_actions.loc[next_action_hour9] = 3.5

    fig_actions = plotting.create_graph(
        df_actions,
        unit="MW",
        title="Ordered balancing actions",
        x_range=x_range,
        y_label="Power (in MW)",
    )
    fig_actions.plot_height = 150
    fig_actions.plot_width = fig_width
    fig_actions.xaxis.visible = False

    if current_user.is_authenticated and (
        current_user.has_role("admin")
        or "wind" in current_user.email
        or "charging" in current_user.email
    ):
        plotting.highlight(
            fig_actions,
            next_action_hour4,
            next_action_hour4 + timedelta(hours=1),
            redirect_to="/control",
        )
    return fig_actions


def mock_flex_action_in_main_figure(fig_profile: Figure):
    # show when user has (possible) actions in order book for a time slot
    if current_user.is_authenticated and (
        current_user.has_role("admin")
        or "wind" in current_user.email
        or "charging" in current_user.email
    ):
        next_action_hour = get_flex_action_hour(4)
        plotting.highlight(
            fig_profile,
            next_action_hour,
            next_action_hour + timedelta(hours=1),
            redirect_to="/control",
        )


def get_flex_action_hour(h: int) -> datetime:
    """ get the next hour from now on """
    this_hour = time_utils.get_most_recent_hour()
    return [
        dt
        for dt in [this_hour + timedelta(hours=i) for i in range(1, 25)]
        if dt.hour == h
    ][0]



from flask import url_for
import pytest

from flexmeasures.data.services.users import find_user_by_email
from flexmeasures.ui.tests.utils import mock_user_response

"""
Testing if the UI crud views do auth checks and display answers.
Actual logic is tested in the API tests.
"""


@pytest.mark.parametrize(
    "view", ["index", "get", "toggle_active", "reset_password_for"]
)
def test_user_crud_as_non_admin(client, as_prosumer, view):
    user_index = client.get(url_for("UserCrudUI:index"), follow_redirects=True)
    assert user_index.status_code == 403
    prosumer2_id = find_user_by_email("test_prosumer2@seita.nl").id
    user_page = client.get(
        url_for(f"UserCrudUI:{view}", id=prosumer2_id), follow_redirects=True
    )
    assert user_page.status_code == 403


def test_user_list(client, as_admin, requests_mock):
    requests_mock.get(
        "http://localhost//api/v2_0/users",
        status_code=200,
        json=mock_user_response(multiple=True),
    )
    user_index = client.get(url_for("UserCrudUI:index"), follow_redirects=True)
    assert user_index.status_code == 200
    assert b"All active users" in user_index.data
    assert b"alex@seita.nl" in user_index.data
    assert b"bert@seita.nl" in user_index.data


def test_user_page(client, as_admin, requests_mock):
    mock_user = mock_user_response(as_list=False)
    requests_mock.get(
        "http://localhost//api/v2_0/user/2", status_code=200, json=mock_user
    )
    requests_mock.get(
        "http://localhost//api/v2_0/assets",
        status_code=200,
        json=[{}, {}, {}],  # we only care about the length
    )
    user_page = client.get(url_for("UserCrudUI:get", id=2), follow_redirects=True)
    assert user_page.status_code == 200
    assert ("Overview for user %s" % mock_user["username"]).encode() in user_page.data
    assert (">3</a>").encode() in user_page.data  # this is the asset count
    assert mock_user["email"].encode() in user_page.data


def test_deactivate_user(client, as_admin, requests_mock):
    """Test it does not fail (logic is tested in API tests) and displays an answer."""
    prosumer2 = find_user_by_email("test_prosumer2@seita.nl", keep_in_session=False)
    requests_mock.patch(
        f"http://localhost//api/v2_0/user/{prosumer2.id}",
        status_code=200,
        json={"active": False},
    )
    # de-activate
    user_page = client.get(
        url_for("UserCrudUI:toggle_active", id=prosumer2.id), follow_redirects=True
    )
    assert user_page.status_code == 200
    assert prosumer2.username in str(user_page.data)
    assert b"new activation status is now False" in user_page.data


def test_reset_password(client, as_admin, requests_mock):
    """Test it does not fail (logic is tested in API tests) and displays an answer."""
    prosumer2 = find_user_by_email("test_prosumer2@seita.nl", keep_in_session=False)
    requests_mock.patch(
        f"http://localhost//api/v2_0/user/{prosumer2.id}/password-reset",
        status_code=200,
    )
    user_page = client.get(
        url_for("UserCrudUI:reset_password_for", id=prosumer2.id),
        follow_redirects=True,
    )
    assert user_page.status_code == 200
    assert b"has been changed to a random password" in user_page.data



from typing import List

import click
from flask import current_app as app
from flask.cli import with_appcontext
from rq import Queue, Worker
from sqlalchemy.orm import configure_mappers

from flexmeasures.data.services.forecasting import handle_forecasting_exception


@click.group("jobs")
def fm_jobs():
    """FlexMeasures: Job queueing."""


@fm_jobs.command("run-worker")
@with_appcontext
@click.option(
    "--name",
    default=None,
    required=True,
    help="Give your worker a recognizable name. Defaults to random string.",
)
@click.option(
    "--queue",
    default=None,
    required=True,
    help="State which queue(s) to work on (using '|' as separator), e.g. 'forecasting', 'scheduling' or 'forecasting|scheduling'.",
)
def run_worker(name: str, queue: str):
    """
    Start a worker process for forecasting and/or scheduling jobs.

    We use the app context to find out which redis queues to use.
    """

    q_list = parse_queue_list(queue)

    # https://stackoverflow.com/questions/50822822/high-sqlalchemy-initialization-overhead
    configure_mappers()

    worker = Worker(
        q_list,
        connection=app.queues["forecasting"].connection,
        name=name,
        exception_handlers=[handle_forecasting_exception],
    )

    click.echo("\n=========================================================")
    click.echo(
        'Worker "%s" initialised: %s (processing %s queues)'
        % (worker.name, worker, len(q_list))
    )
    for q in q_list:
        click.echo("Running against %s on %s" % (q, q.connection))
    click.echo("=========================================================\n")

    worker.work()


@fm_jobs.command("clear-queue")
@with_appcontext
@click.option(
    "--queue",
    default=None,
    required=True,
    help="State which queue(s) to clear (using '|' as separator), e.g. 'forecasting', 'scheduling' or 'forecasting|scheduling'.",
)
def clear_queue(queue: str):
    """
    Clear a job queue.

    We use the app context to find out which redis queues to use.
    """

    q_list = parse_queue_list(queue)
    configure_mappers()
    for q in q_list:
        count_before = q.count
        q.empty()
        count_after = q.count
        click.echo(
            f"Cleared {count_before - count_after} jobs from {q}. Queue now contains {count_after} jobs."
        )


def parse_queue_list(queue_names_str: str) -> List[Queue]:
    """Parse a | separated string of queue names against the app.queues dict.

    The app.queues dict is expected to have queue names as keys, and rq.Queue objects as values.

    :param queue_names_str: a string with queue names separated by the | character
    :returns:               a list of Queue objects.
    """
    q_list = []
    for q_name in queue_names_str.split("|"):
        if q_name in app.queues:
            q_list.append(app.queues[q_name])
        else:
            raise ValueError(f"Unknown queue '{q_name}'.")
    return q_list


app.cli.add_command(fm_jobs)



"""sensor.generic_asset_id should not be orphaned on delete

Revision ID: e4c9cf837311
Revises: 994170c26bc6
Create Date: 2021-08-15 20:42:29.729532

"""
from alembic import op


# revision identifiers, used by Alembic.
revision = "e4c9cf837311"
down_revision = "994170c26bc6"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_constraint(
        "sensor_generic_asset_id_generic_asset_fkey", "sensor", type_="foreignkey"
    )
    op.create_foreign_key(
        op.f("sensor_generic_asset_id_generic_asset_fkey"),
        "sensor",
        "generic_asset",
        ["generic_asset_id"],
        ["id"],
        ondelete="CASCADE",
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_constraint(
        op.f("sensor_generic_asset_id_generic_asset_fkey"), "sensor", type_="foreignkey"
    )
    op.create_foreign_key(
        "sensor_generic_asset_id_generic_asset_fkey",
        "sensor",
        "generic_asset",
        ["generic_asset_id"],
        ["id"],
    )
    # ### end Alembic commands ###



from typing import List
import pytest
from datetime import timedelta

import isodate
from flask_security import SQLAlchemySessionUserDatastore
from flask_security.utils import hash_password

from flexmeasures.data.models.assets import Power
from flexmeasures.data.models.data_sources import DataSource
from flexmeasures.data.services.users import create_user


@pytest.fixture(scope="module")
def setup_api_test_data(db, setup_account, setup_roles_users, add_market_prices):
    """
    Set up data for API v1.1 tests.
    """
    print("Setting up data for API v1.1 tests on %s" % db.engine)

    from flexmeasures.data.models.user import User, Role
    from flexmeasures.data.models.assets import Asset, AssetType

    user_datastore = SQLAlchemySessionUserDatastore(db.session, User, Role)

    # Create a user without proper registration as a data source
    user = user_datastore.create_user(
        username="test user with improper registration",
        email="test_improper_user@seita.nl",
        password=hash_password("testtest"),
        account_id=setup_account.id,
    )
    role = user_datastore.find_role("Prosumer")
    user_datastore.add_role_to_user(user, role)

    # Create a test user without a USEF role
    create_user(
        username="test user without roles",
        email="test_user@seita.nl",
        password=hash_password("testtest"),
        account_name=setup_account.name,
    )

    # Create 3 test assets for the test_prosumer user
    test_prosumer = setup_roles_users["Test Prosumer"]
    test_asset_type = AssetType(name="test-type")
    db.session.add(test_asset_type)
    asset_names = ["CS 1", "CS 2", "CS 3"]
    assets: List[Asset] = []
    for asset_name in asset_names:
        asset = Asset(
            name=asset_name,
            asset_type_name="test-type",
            event_resolution=timedelta(minutes=15),
            capacity_in_mw=1,
            latitude=100,
            longitude=100,
            unit="MW",
        )
        asset.owner = test_prosumer
        assets.append(asset)
        db.session.add(asset)

    # Add power forecasts to the assets
    cs_1 = Asset.query.filter(Asset.name == "CS 1").one_or_none()
    cs_2 = Asset.query.filter(Asset.name == "CS 2").one_or_none()
    cs_3 = Asset.query.filter(Asset.name == "CS 3").one_or_none()
    data_source = DataSource.query.filter(
        DataSource.user == test_prosumer
    ).one_or_none()
    power_forecasts = []
    for i in range(6):
        p_1 = Power(
            datetime=isodate.parse_datetime("2015-01-01T00:00:00Z")
            + timedelta(minutes=15 * i),
            horizon=timedelta(hours=6),
            value=(300 + i) * -1,
            asset_id=cs_1.id,
            data_source_id=data_source.id,
        )
        p_2 = Power(
            datetime=isodate.parse_datetime("2015-01-01T00:00:00Z")
            + timedelta(minutes=15 * i),
            horizon=timedelta(hours=6),
            value=(300 - i) * -1,
            asset_id=cs_2.id,
            data_source_id=data_source.id,
        )
        p_3 = Power(
            datetime=isodate.parse_datetime("2015-01-01T00:00:00Z")
            + timedelta(minutes=15 * i),
            horizon=timedelta(hours=6),
            value=(0 + i) * -1,
            asset_id=cs_3.id,
            data_source_id=data_source.id,
        )
        power_forecasts.append(p_1)
        power_forecasts.append(p_2)
        power_forecasts.append(p_3)
    db.session.bulk_save_objects(power_forecasts)

    print("Done setting up data for API v1.1 tests")


@pytest.fixture(scope="function")
def setup_fresh_api_v1_1_test_data(
    fresh_db, setup_roles_users_fresh_db, setup_markets_fresh_db
):
    return fresh_db



from datetime import datetime

from flask_security import UserMixin, RoleMixin
from sqlalchemy.orm import relationship, backref
from sqlalchemy import Boolean, DateTime, Column, Integer, String, ForeignKey
from sqlalchemy.ext.hybrid import hybrid_property

from flexmeasures.data.config import db


class RolesAccounts(db.Model):
    __tablename__ = "roles_accounts"
    id = Column(Integer(), primary_key=True)
    account_id = Column("account_id", Integer(), ForeignKey("account.id"))
    role_id = Column("role_id", Integer(), ForeignKey("account_role.id"))


class AccountRole(db.Model):
    __tablename__ = "account_role"
    id = Column(Integer(), primary_key=True)
    name = Column(String(80), unique=True)
    description = Column(String(255))

    def __repr__(self):
        return "<AccountRole:%s (ID:%d)>" % (self.name, self.id)


class Account(db.Model):
    """
    Account of a tenant on the server.
    Bundles Users as well as GenericAssets.
    """

    __tablename__ = "account"
    id = Column(Integer, primary_key=True)
    name = Column(String(100), default="", unique=True)
    account_roles = relationship(
        "AccountRole",
        secondary="roles_accounts",
        backref=backref("accounts", lazy="dynamic"),
    )

    def __repr__(self):
        return "<Account %s (ID:%d)" % (self.name, self.id)


class RolesUsers(db.Model):
    __tablename__ = "roles_users"
    id = Column(Integer(), primary_key=True)
    user_id = Column("user_id", Integer(), ForeignKey("fm_user.id"))
    role_id = Column("role_id", Integer(), ForeignKey("role.id"))


class Role(db.Model, RoleMixin):
    __tablename__ = "role"
    id = Column(Integer(), primary_key=True)
    name = Column(String(80), unique=True)
    description = Column(String(255))

    def __repr__(self):
        return "<Role:%s (ID:%d)>" % (self.name, self.id)


class User(db.Model, UserMixin):
    """
    We use the flask security UserMixin, which does include functionality,
    but not the fields (those are in flask_security/models::FsUserMixin).
    We went with a pick&choose approach. This gives us more freedom, e.g.
    to choose our own table name or add logic around the activation status.
    If we add new FS functionality (e.g. 2FA), the fields needed for that
    need to be added here.
    """

    __tablename__ = "fm_user"
    id = Column(Integer, primary_key=True)
    email = Column(String(255), unique=True)
    username = Column(String(255), unique=True)
    password = Column(String(255))
    last_login_at = Column(DateTime())
    login_count = Column(Integer)
    active = Column(Boolean())
    # Faster token checking
    fs_uniquifier = Column(String(64), unique=True, nullable=False)
    timezone = Column(String(255), default="Europe/Amsterdam")
    account_id = Column(Integer, db.ForeignKey("account.id"), nullable=False)

    account = db.relationship("Account", backref=db.backref("users", lazy=True))
    flexmeasures_roles = relationship(
        "Role",
        secondary="roles_users",
        backref=backref("users", lazy="dynamic"),
    )

    def __repr__(self):
        return "<User %s (ID:%d)" % (self.username, self.id)

    @property
    def is_authenticated(self):
        """We are overloading this, so it also considers being active.
        Inactive users can by definition not be authenticated."""
        return super(UserMixin, self).is_authenticated and self.active

    @hybrid_property
    def roles(self):
        """The roles attribute is being used by Flask-Security in the @roles_required decorator (among others).
        With this little overload fix, it will only return the user's roles if they are authenticated.
        We do this to prevent that if a user is logged in while the admin deactivates them, their session would still work.
        In effect, we strip unauthenticated users from their roles. To read roles of an unauthenticated user
        (e.g. being inactive), use the `flexmeasures_roles` attribute.
        If our auth model has moved to an improved way, e.g. requiring modern tokens, we should consider relaxing this.
        Note: This needed to become a hybrid property when moving to Flask-Security 3.4
        """
        if not self.is_authenticated and self is not User:
            return []
        else:
            return self.flexmeasures_roles

    @roles.setter
    def roles(self, new_roles):
        """See comment in roles property why we overload."""
        self.flexmeasures_roles = new_roles

    def has_role(self, role):
        """Returns `True` if the user identifies with the specified role.
            Overwritten from flask_security.core.UserMixin.

        :param role: A role name or `Role` instance"""
        if isinstance(role, str):
            return role in (role.name for role in self.flexmeasures_roles)
        else:
            return role in self.flexmeasures_roles


def remember_login(the_app, user):
    """We do not use the tracking feature of flask_security, but this basic meta data are quite handy to know"""
    user.last_login_at = datetime.utcnow()
    if user.login_count is None:
        user.login_count = 0
    user.login_count = user.login_count + 1


def is_user(o) -> bool:
    """True if object is or proxies a User, False otherwise.

    Takes into account that object can be of LocalProxy type, and
    uses get_current_object to get the underlying (User) object.
    """
    return isinstance(o, User) or (
        hasattr(o, "_get_current_object") and isinstance(o._get_current_object(), User)
    )



"""merge

Revision ID: 919dc9f1dc1f
Revises: db00b66be82c, 5d39829d91af
Create Date: 2018-07-15 22:49:59.184453

"""


# revision identifiers, used by Alembic.
revision = "919dc9f1dc1f"
down_revision = ("db00b66be82c", "5d39829d91af")
branch_labels = None
depends_on = None


def upgrade():
    pass


def downgrade():
    pass



from typing import Union, Type

from flexmeasures.data.models.assets import AssetType, Asset, Power
from flexmeasures.data.models.markets import MarketType, Market, Price
from flexmeasures.data.models.weather import WeatherSensorType, WeatherSensor, Weather


def determine_asset_type_by_asset(
    generic_asset: Union[Asset, Market, WeatherSensor]
) -> Union[AssetType, MarketType, WeatherSensorType]:
    if isinstance(generic_asset, Asset):
        return generic_asset.asset_type
    elif isinstance(generic_asset, Market):
        return generic_asset.market_type
    elif isinstance(generic_asset, WeatherSensor):
        return generic_asset.sensor_type
    else:
        raise TypeError("Unknown generic asset type.")


def determine_asset_value_class_by_asset(
    generic_asset: Union[Asset, Market, WeatherSensor]
) -> Type[Union[Power, Price, Weather]]:
    if isinstance(generic_asset, Asset):
        return Power
    elif isinstance(generic_asset, Market):
        return Price
    elif isinstance(generic_asset, WeatherSensor):
        return Weather
    else:
        raise TypeError("Unknown generic asset type.")



from typing import Optional

from statsmodels.api import OLS

from flexmeasures.data.models.forecasting.model_spec_factory import (
    create_initial_model_specs,
)

"""
Simple linear regression by ordinary least squares.
"""

# update this version if small things like parametrisation change
version: int = 2
# if a forecasting job using this model fails, fall back on this one
fallback_model_search_term: Optional[str] = "naive"


def ols_specs_configurator(**kwargs):
    """Create and customize initial specs with OLS. See model_spec_factory for param docs."""
    model_specs = create_initial_model_specs(**kwargs)
    model_specs.set_model(OLS)
    model_identifier = "linear-OLS model v%d" % version
    return model_specs, model_identifier, fallback_model_search_term



"""make_data_source_id_columns_primary_keys

Revision ID: 8fcc5fec67bc
Revises: 9c7fc8e46f1e
Create Date: 2018-07-30 15:39:30.583000

"""
from alembic import op


# revision identifiers, used by Alembic.
revision = "8fcc5fec67bc"
down_revision = "9c7fc8e46f1e"
branch_labels = None
depends_on = None

"""
This revision adds the data_source_id field to the primary key of time series tables.
"""


def upgrade():

    # Power
    op.execute("alter table power drop constraint power_pkey;")
    op.execute(
        "alter table power add constraint power_pkey primary key (datetime, asset_id, horizon, data_source_id);"
    )
    # Price
    op.execute("alter table price drop constraint price_pkey;")
    op.execute(
        "alter table price add constraint price_pkey primary key (datetime, market_id, horizon, data_source_id);"
    )
    # Weather
    op.execute("alter table weather drop constraint weather_pkey;")
    op.execute(
        "alter table weather add constraint weather_pkey primary key (datetime, sensor_id, horizon, data_source_id);"
    )


def downgrade():

    # Power
    op.execute("alter table power drop constraint power_pkey;")
    op.execute(
        "alter table power add constraint power_pkey primary key (datetime, asset_id, horizon);"
    )
    # Price
    op.execute("alter table price drop constraint price_pkey;")
    op.execute(
        "alter table price add constraint price_pkey primary key (datetime, market_id, horizon;"
    )
    # Weather
    op.execute("alter table weather drop constraint weather_pkey;")
    op.execute(
        "alter table weather add constraint weather_pkey primary key (datetime, sensor_id, horizon);"
    )



from typing import Optional, Tuple

from flexmeasures.data import db


class GenericAssetType(db.Model):
    """An asset type defines what type an asset belongs to.

    Examples of asset types: WeatherStation, Market, CP, EVSE, WindTurbine, SolarPanel, Building.
    """

    id = db.Column(db.Integer, primary_key=True)
    name = db.Column(db.String(80), default="")
    description = db.Column(db.String(80), nullable=True, unique=False)


class GenericAsset(db.Model):
    """An asset is something that has economic value.

    Examples of tangible assets: a house, a ship, a weather station.
    Examples of intangible assets: a market, a country, a copyright.
    """

    id = db.Column(db.Integer, primary_key=True)
    name = db.Column(db.String(80), default="")
    latitude = db.Column(db.Float, nullable=True)
    longitude = db.Column(db.Float, nullable=True)

    generic_asset_type_id = db.Column(
        db.Integer, db.ForeignKey("generic_asset_type.id"), nullable=False
    )
    generic_asset_type = db.relationship(
        "GenericAssetType",
        foreign_keys=[generic_asset_type_id],
        backref=db.backref("generic_assets", lazy=True),
    )

    account_id = db.Column(
        db.Integer, db.ForeignKey("account.id", ondelete="CASCADE"), nullable=True
    )  # if null, asset is public

    owner = db.relationship(
        "Account",
        backref=db.backref(
            "generic_assets",
            foreign_keys=[account_id],
            lazy=True,
            cascade="all, delete-orphan",
            passive_deletes=True,
        ),
    )

    @property
    def location(self) -> Optional[Tuple[float, float]]:
        if None not in (self.latitude, self.longitude):
            return self.latitude, self.longitude
        return None


def create_generic_asset(generic_asset_type: str, **kwargs) -> GenericAsset:
    """Create a GenericAsset and assigns it an id.

    :param generic_asset_type: "asset", "market" or "weather_sensor"
    :param kwargs:              should have values for keys "name", and:
                                - "asset_type_name" or "asset_type" when generic_asset_type is "asset"
                                - "market_type_name" or "market_type" when generic_asset_type is "market"
                                - "weather_sensor_type_name" or "weather_sensor_type" when generic_asset_type is "weather_sensor"
                                - alternatively, "sensor_type" is also fine
    :returns:                   the created GenericAsset
    """
    asset_type_name = kwargs.pop(f"{generic_asset_type}_type_name", None)
    if asset_type_name is None:
        if f"{generic_asset_type}_type" in kwargs:
            asset_type_name = kwargs.pop(f"{generic_asset_type}_type").name
        else:
            asset_type_name = kwargs.pop("sensor_type").name
    generic_asset_type = GenericAssetType.query.filter_by(
        name=asset_type_name
    ).one_or_none()
    if generic_asset_type is None:
        raise ValueError(f"Cannot find GenericAssetType {asset_type_name} in database.")
    new_generic_asset = GenericAsset(
        name=kwargs["name"], generic_asset_type_id=generic_asset_type.id
    )
    for arg in ("latitude", "longitude", "account_id"):
        if arg in kwargs:
            setattr(new_generic_asset, arg, kwargs[arg])
    db.session.add(new_generic_asset)
    db.session.flush()  # generates the pkey for new_generic_asset
    return new_generic_asset



from flask_security import auth_token_required

from flexmeasures.api.common.utils.api_utils import list_access
from flexmeasures.api.common.utils.decorators import as_response_type
from flexmeasures.api.common.utils.validators import usef_roles_accepted
from flexmeasures.api.v1 import (
    flexmeasures_api as flexmeasures_api_v1,
    implementations as v1_implementations,
)


# The service listing for this API version (import from previous version or update if needed)
v1_service_listing = {
    "version": "1.0",
    "services": [
        {
            "name": "getMeterData",
            "access": ["Aggregator", "Supplier", "MDC", "DSO", "Prosumer", "ESCo"],
            "description": "Request meter reading",
        },
        {
            "name": "postMeterData",
            "access": ["Aggregator", "Supplier", "MDC", "DSO", "Prosumer", "ESCo"],
            "description": "Send meter reading",
        },
    ],
}


@flexmeasures_api_v1.route("/getMeterData", methods=["GET", "POST"])
@as_response_type("GetMeterDataResponse")
@auth_token_required
@usef_roles_accepted(*list_access(v1_service_listing, "getMeterData"))
def get_meter_data():
    """API endpoint to get meter data.

    .. :quickref: Data; Download meter data from the platform

    **Optional fields**

    - "resolution" (see :ref:`resolutions`)
    - "horizon" (see :ref:`beliefs`)
    - "prior" (see :ref:`beliefs`)
    - "source" (see :ref:`sources`)

    **Example request**

    This "GetMeterDataRequest" message requests measured consumption between 0.00am and 1.30am for charging station 1.

    .. code-block:: json

        {
            "type": "GetMeterDataRequest",
            "connection": "CS 1",
            "start": "2015-01-01T00:00:00Z",
            "duration": "PT1H30M",
            "unit": "MW"
        }

    **Example response**

    This "GetMeterDataResponse" message indicates that consumption for charging station 1 was measured in 15-minute
    intervals.

    .. sourcecode:: json

        {
            "type": "GetMeterDataResponse",
            "connection": "CS 1",
            "values": [
                306.66,
                306.66,
                0,
                0,
                306.66,
                306.66
            ],
            "start": "2015-01-01T00:00:00Z",
            "duration": "PT1H30M",
            "unit": "MW"
        }

    :reqheader Authorization: The authentication token
    :reqheader Content-Type: application/json
    :resheader Content-Type: application/json
    :status 200: PROCESSED
    :status 400: INVALID_DOMAIN, INVALID_MESSAGE_TYPE, INVALID_SOURCE, INVALID_TIMEZONE, INVALID_UNIT, UNRECOGNIZED_ASSET, or UNRECOGNIZED_CONNECTION_GROUP
    :status 401: UNAUTHORIZED
    :status 403: INVALID_SENDER
    :status 405: INVALID_METHOD
    """
    return v1_implementations.get_meter_data_response()


@flexmeasures_api_v1.route("/postMeterData", methods=["POST"])
@as_response_type("PostMeterDataResponse")
@auth_token_required
@usef_roles_accepted(*list_access(v1_service_listing, "postMeterData"))
def post_meter_data():
    """API endpoint to post meter data.

    .. :quickref: Data; Upload meter data to the platform

    **Optional fields**

    - "horizon" (see :ref:`prognoses`)

    **Example request**

    This "PostMeterDataRequest" message posts measured consumption for 15-minute intervals between 0.00am and 1.30am for
    charging stations 1, 2 and 3 (negative values denote production).

    .. code-block:: json

        {
            "type": "PostMeterDataRequest",
            "groups": [
                {
                    "connections": [
                        "CS 1",
                        "CS 3"
                    ],
                    "values": [
                        306.66,
                        306.66,
                        0,
                        0,
                        306.66,
                        306.66
                    ]
                },
                {
                    "connections": [
                        "CS 2"
                    ],
                    "values": [
                        306.66,
                        0,
                        0,
                        0,
                        306.66,
                        306.66
                    ]
                }
            ],
            "start": "2015-01-01T00:00:00Z",
            "duration": "PT1H30M",
            "unit": "MW"
        }

    It is allowed to send higher resolutions (in this example for instance, 30 minutes) which will be upsampled.

    **Example response**

    This "PostMeterDataResponse" message indicates that the measurement has been processed without any error.

    .. sourcecode:: json

        {
            "type": "PostMeterDataResponse",
            "status": "PROCESSED",
            "message": "Request has been processed."
        }

    :reqheader Authorization: The authentication token
    :reqheader Content-Type: application/json
    :resheader Content-Type: application/json
    :status 200: PROCESSED
    :status 400: INVALID_DOMAIN, INVALID_MESSAGE_TYPE, INVALID_TIMEZONE, INVALID_UNIT, REQUIRED_INFO_MISSING, UNRECOGNIZED_ASSET or UNRECOGNIZED_CONNECTION_GROUP
    :status 401: UNAUTHORIZED
    :status 403: INVALID_SENDER
    :status 405: INVALID_METHOD
    """
    return v1_implementations.post_meter_data_response()


@flexmeasures_api_v1.route("/getService", methods=["GET"])
@as_response_type("GetServiceResponse")
def get_service():
    """API endpoint to get a service listing for this version.

    .. :quickref: Public; Obtain a service listing for this version

    :resheader Content-Type: application/json
    :status 200: PROCESSED
    """
    return v1_implementations.get_service_response(v1_service_listing)



from datetime import timedelta

from flask_security import SQLAlchemySessionUserDatastore
import pytest

from flexmeasures.data.models.generic_assets import GenericAssetType, GenericAsset
from flexmeasures.data.models.time_series import Sensor


@pytest.fixture(scope="module", autouse=True)
def setup_api_test_data(db, setup_roles_users):
    """
    Set up data for API dev tests.
    """
    print("Setting up data for API v2.0 tests on %s" % db.engine)
    add_gas_sensor(db, setup_roles_users["Test Supplier"])
    give_prosumer_the_MDC_role(db)


@pytest.fixture(scope="function")
def setup_api_fresh_test_data(fresh_db, setup_roles_users_fresh_db):
    """
    Set up fresh data for API dev tests.
    """
    print("Setting up fresh data for API dev tests on %s" % fresh_db.engine)
    for sensor in Sensor.query.all():
        fresh_db.delete(sensor)
    add_gas_sensor(fresh_db, setup_roles_users_fresh_db["Test Supplier"])
    give_prosumer_the_MDC_role(fresh_db)


def add_gas_sensor(db, test_supplier):
    incineration_type = GenericAssetType(
        name="waste incinerator",
    )
    db.session.add(incineration_type)
    db.session.flush()
    incineration_asset = GenericAsset(
        name="incineration line",
        generic_asset_type=incineration_type,
    )
    db.session.add(incineration_asset)
    db.session.flush()
    gas_sensor = Sensor(
        name="some gas sensor",
        unit="m³/h",
        event_resolution=timedelta(minutes=10),
        generic_asset=incineration_asset,
    )
    db.session.add(gas_sensor)
    gas_sensor.owner = test_supplier


def give_prosumer_the_MDC_role(db):

    from flexmeasures.data.models.user import User, Role

    user_datastore = SQLAlchemySessionUserDatastore(db.session, User, Role)
    test_prosumer = user_datastore.find_user(email="test_prosumer@seita.nl")
    mdc_role = user_datastore.create_role(name="MDC", description="Meter Data Company")
    user_datastore.add_role_to_user(test_prosumer, mdc_role)



from datetime import timedelta

from flask import url_for
import isodate
import pandas as pd
import pytest

from flexmeasures.api.common.responses import (
    invalid_domain,
    invalid_sender,
    invalid_unit,
    request_processed,
    unrecognized_connection_group,
)
from flexmeasures.api.tests.utils import get_auth_token
from flexmeasures.api.common.utils.api_utils import message_replace_name_with_ea
from flexmeasures.api.common.utils.validators import validate_user_sources
from flexmeasures.api.v1.tests.utils import (
    message_for_get_meter_data,
    message_for_post_meter_data,
    verify_power_in_db,
)
from flexmeasures.data.auth_setup import UNAUTH_ERROR_STATUS
from flexmeasures.data.models.assets import Asset


@pytest.mark.parametrize("query", [{}, {"access": "Prosumer"}])
def test_get_service(client, query):
    get_service_response = client.get(
        url_for("flexmeasures_api_v1.get_service"),
        query_string=query,
        headers={"content-type": "application/json"},
    )
    print("Server responded with:\n%s" % get_service_response.json)
    assert get_service_response.status_code == 200
    assert get_service_response.json["type"] == "GetServiceResponse"
    assert get_service_response.json["status"] == request_processed()[0]["status"]
    if "access" in query:
        for service in get_service_response.json["services"]:
            assert "Prosumer" in service["access"]


def test_unauthorized_request(client):
    get_meter_data_response = client.get(
        url_for("flexmeasures_api_v1.get_meter_data"),
        query_string=message_for_get_meter_data(no_connection=True),
        headers={"content-type": "application/json"},
    )
    print("Server responded with:\n%s" % get_meter_data_response.json)
    assert get_meter_data_response.status_code == 401
    assert get_meter_data_response.json["type"] == "GetMeterDataResponse"
    assert get_meter_data_response.json["status"] == UNAUTH_ERROR_STATUS


def test_no_connection_in_get_request(client):
    get_meter_data_response = client.get(
        url_for("flexmeasures_api_v1.get_meter_data"),
        query_string=message_for_get_meter_data(no_connection=True),
        headers={
            "Authorization": get_auth_token(
                client, "test_prosumer@seita.nl", "testtest"
            )
        },
    )
    print("Server responded with:\n%s" % get_meter_data_response.json)
    assert get_meter_data_response.status_code == 400
    assert get_meter_data_response.json["type"] == "GetMeterDataResponse"
    assert (
        get_meter_data_response.json["status"]
        == unrecognized_connection_group()[0]["status"]
    )


def test_invalid_connection_in_get_request(client):
    get_meter_data_response = client.get(
        url_for("flexmeasures_api_v1.get_meter_data"),
        query_string=message_for_get_meter_data(invalid_connection=True),
        headers={
            "Authorization": get_auth_token(
                client, "test_prosumer@seita.nl", "testtest"
            )
        },
    )
    print("Server responded with:\n%s" % get_meter_data_response.json)
    assert get_meter_data_response.status_code == 400
    assert get_meter_data_response.json["type"] == "GetMeterDataResponse"
    assert get_meter_data_response.json["status"] == invalid_domain()[0]["status"]


@pytest.mark.parametrize("method", ["GET", "POST"])
@pytest.mark.parametrize(
    "message",
    [
        message_for_get_meter_data(no_unit=True),
        message_for_get_meter_data(invalid_unit=True),
    ],
)
def test_invalid_or_no_unit(client, method, message):
    if method == "GET":
        get_meter_data_response = client.get(
            url_for("flexmeasures_api_v1.get_meter_data"),
            query_string=message,
            headers={
                "Authorization": get_auth_token(
                    client, "test_prosumer@seita.nl", "testtest"
                )
            },
        )
    elif method == "POST":
        get_meter_data_response = client.post(
            url_for("flexmeasures_api_v1.get_meter_data"),
            json=message,
            headers={
                "Authorization": get_auth_token(
                    client, "test_prosumer@seita.nl", "testtest"
                )
            },
        )
    else:
        get_meter_data_response = []
    assert get_meter_data_response.status_code == 400
    assert get_meter_data_response.json["type"] == "GetMeterDataResponse"
    assert (
        get_meter_data_response.json["status"]
        == invalid_unit("power", ["MW"])[0]["status"]
    )


@pytest.mark.parametrize(
    "user_email, get_message",
    [
        ["test_user@seita.nl", message_for_get_meter_data()],
        ["demo@seita.nl", message_for_get_meter_data(demo_connection=True)],
    ],
)
def test_invalid_sender_and_logout(client, user_email, get_message):
    """
    Tries to get meter data as a logged-in test user without any USEF role, which should fail.
    Then tries to log out, which should succeed as a url direction.
    """

    # get meter data
    auth_token = get_auth_token(client, user_email, "testtest")
    get_meter_data_response = client.get(
        url_for("flexmeasures_api_v1.get_meter_data"),
        query_string=get_message,
        headers={"Authorization": auth_token},
    )
    print("Server responded with:\n%s" % get_meter_data_response.json)
    assert get_meter_data_response.status_code == 403
    assert get_meter_data_response.json["type"] == "GetMeterDataResponse"
    assert get_meter_data_response.json["status"] == invalid_sender("MDC")[0]["status"]

    # log out
    logout_response = client.get(
        url_for("security.logout"),
        headers={"Authorization ": auth_token, "content-type": "application/json"},
    )
    assert logout_response.status_code == 302


def test_invalid_resolution_str(client):
    auth_token = get_auth_token(client, "test_prosumer@seita.nl", "testtest")
    query_string = message_for_get_meter_data()
    query_string["resolution"] = "15M"  # invalid
    get_meter_data_response = client.get(
        url_for("flexmeasures_api_v1.get_meter_data"),
        query_string=query_string,
        headers={"Authorization": auth_token},
    )
    print("Server responded with:\n%s" % get_meter_data_response.json)
    assert get_meter_data_response.status_code == 400
    assert get_meter_data_response.json["type"] == "GetMeterDataResponse"
    assert get_meter_data_response.json["status"] == "INVALID_RESOLUTION"


@pytest.mark.parametrize(
    "message",
    [
        message_for_get_meter_data(single_connection=True),
        message_for_get_meter_data(
            single_connection=True, source="Prosumer"
        ),  # sourced by a Prosumer
        message_for_get_meter_data(
            single_connection=True, source=["Prosumer", 109304]
        ),  # sourced by a Prosumer or user 109304
    ],
)
def test_get_meter_data(db, app, client, message):
    """Checks Charging Station 5, which has multi-sourced data for the same time interval:
    6 values from a Prosumer, and 6 values from a Supplier.

    All data should be in the database, and currently only the Prosumer data is returned.
    """
    message["connection"] = "CS 5"

    # set up frame with expected values, and filter by source if needed
    expected_values = pd.concat(
        [
            pd.DataFrame.from_dict(
                dict(
                    value=[(100.0 + i) for i in range(6)],
                    datetime=[
                        isodate.parse_datetime("2015-01-01T00:00:00Z")
                        + timedelta(minutes=15 * i)
                        for i in range(6)
                    ],
                    data_source_id=1,
                )
            ),
            pd.DataFrame.from_dict(
                dict(
                    value=[(1000.0 - 10 * i) for i in range(6)],
                    datetime=[
                        isodate.parse_datetime("2015-01-01T00:00:00Z")
                        + timedelta(minutes=15 * i)
                        for i in range(6)
                    ],
                    data_source_id=2,
                )
            ),
        ]
    )
    if "source" in message:
        source_ids = validate_user_sources(message["source"])
        expected_values = expected_values[
            expected_values["data_source_id"].isin(source_ids)
        ]
    expected_values = expected_values.set_index(
        ["datetime", "data_source_id"]
    ).sort_index()

    # check whether conftest.py did its job setting up the database with expected values
    cs_5 = Asset.query.filter(Asset.name == "CS 5").one_or_none()
    verify_power_in_db(message, cs_5, expected_values, db, swapped_sign=True)

    # check whether the API returns the expected values (currently only the Prosumer data is returned)
    auth_token = get_auth_token(client, "test_prosumer@seita.nl", "testtest")
    get_meter_data_response = client.get(
        url_for("flexmeasures_api_v1.get_meter_data"),
        query_string=message_replace_name_with_ea(message),
        headers={"content-type": "application/json", "Authorization": auth_token},
    )
    print("Server responded with:\n%s" % get_meter_data_response.json)
    assert get_meter_data_response.status_code == 200
    assert get_meter_data_response.json["values"] == [(100.0 + i) for i in range(6)]


def test_post_meter_data_to_different_resolutions(app, client):
    """
    Tries to post meter data to assets with different event_resolutions, which is not accepted.
    """

    post_message = message_for_post_meter_data(different_target_resolutions=True)
    auth_token = get_auth_token(client, "test_prosumer@seita.nl", "testtest")
    post_meter_data_response = client.post(
        url_for("flexmeasures_api_v1.post_meter_data"),
        json=message_replace_name_with_ea(post_message),
        headers={"Authorization": auth_token},
    )
    print("Server responded with:\n%s" % post_meter_data_response.json)
    assert post_meter_data_response.json["type"] == "PostMeterDataResponse"
    assert post_meter_data_response.status_code == 400
    assert (
        "assets do not have matching resolutions"
        in post_meter_data_response.json["message"]
    )
    assert "CS 2" in post_meter_data_response.json["message"]
    assert "CS 4" in post_meter_data_response.json["message"]
    assert post_meter_data_response.json["status"] == "INVALID_RESOLUTION"



"""Add forecasting_job table

Revision ID: 1b64acf01809
Revises: 1a4f0e5c4b86
Create Date: 2018-08-31 15:09:09.462860

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = "1b64acf01809"
down_revision = "1a4f0e5c4b86"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "forecasting_job",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("timed_value_type", sa.String(length=30), nullable=False),
        sa.Column("asset_id", sa.Integer(), nullable=False),
        sa.Column("start", sa.DateTime(timezone=True), nullable=False),
        sa.Column("end", sa.DateTime(timezone=True), nullable=False),
        sa.Column("horizon", sa.Interval(), nullable=False),
        sa.Column("in_progress_since", sa.DateTime(timezone=True), nullable=True),
        sa.PrimaryKeyConstraint("id"),
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table("forecasting_job")
    # ### end Alembic commands ###



"""weather sensor unique type and location

Revision ID: b2b43f0eec40
Revises: 1b64acf01809
Create Date: 2018-09-12 11:14:46.486640

"""
from alembic import op


# revision identifiers, used by Alembic.
revision = "b2b43f0eec40"
down_revision = "1b64acf01809"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_unique_constraint(
        "_type_name_location_unique",
        "weather_sensor",
        ["weather_sensor_type_name", "latitude", "longitude"],
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_constraint("_type_name_location_unique", "weather_sensor", type_="unique")
    # ### end Alembic commands ###



"""unique usernames

Revision ID: 1bcccdf0c3e1
Revises: 8fcc5fec67bc
Create Date: 2018-08-08 13:39:56.732107

"""
from alembic import op


# revision identifiers, used by Alembic.
revision = "1bcccdf0c3e1"
down_revision = "8fcc5fec67bc"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_unique_constraint(None, "bvp_users", ["username"])
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_constraint(None, "bvp_users", type_="unique")
    # ### end Alembic commands ###



from datetime import datetime, timedelta

from flask import url_for
import pytz
import isodate

from flexmeasures.api.tests.utils import get_auth_token, get_task_run, post_task_run
from flexmeasures.data.auth_setup import (
    FORBIDDEN_ERROR_STATUS,
    FORBIDDEN_STATUS_CODE,
    FORBIDDEN_ERROR_CLASS,
    UNAUTH_STATUS_CODE,
)


def test_api_task_run_post_unauthorized_wrong_role(client):
    url = url_for("flexmeasures_api_ops.post_task_run")
    auth_token = get_auth_token(client, "test_prosumer@seita.nl", "testtest")
    post_req_params = dict(
        query_string={"name": "my-task"}, headers={"Authorization": auth_token}
    )
    task_run = client.post(url, **post_req_params)
    assert task_run.status_code == FORBIDDEN_STATUS_CODE
    assert bytes(FORBIDDEN_ERROR_CLASS, encoding="utf") in task_run.data
    # While we are on it, test if the unauth handler correctly returns json if we set the content-type
    post_req_params.update(
        headers={"Authorization": auth_token, "Content-Type": "application/json"}
    )
    task_run = client.post(url, **post_req_params)
    assert task_run.status_code == FORBIDDEN_STATUS_CODE
    assert task_run.json["status"] == FORBIDDEN_ERROR_STATUS


def test_api_task_run_get_no_token(client):
    task_run = get_task_run(client, "task-B", "NOAUTH")
    assert task_run.status_code == UNAUTH_STATUS_CODE
    assert task_run.json["status"] == "ERROR"
    assert "Not authenticated" in task_run.json["reason"]


def test_api_task_run_get_bad_token(client):
    task_run = get_task_run(client, "task-B", "bad-token")
    assert task_run.status_code == FORBIDDEN_STATUS_CODE
    assert task_run.json["status"] == "ERROR"
    assert "Not authorized" in task_run.json["reason"]


def test_api_task_run_get_no_name(client):
    task_run = get_task_run(client, "")
    assert task_run.status_code == 400
    assert task_run.json["status"] == "ERROR"
    assert task_run.json["reason"] == "No task name given."


def test_api_task_run_get_nonexistent_task(client):
    task_run = get_task_run(client, "task-Z")
    assert task_run.status_code == 404
    assert task_run.json["status"] == "ERROR"
    assert "has no last run time" in task_run.json["reason"]


def test_api_task_run_post_no_name(client):
    task_run = post_task_run(client, "")
    assert task_run.status_code == 400
    assert task_run.json["status"] == "ERROR"
    assert task_run.json["reason"] == "No task name given."


def test_api_task_run_get_recent_entry(client):
    task_run = get_task_run(client, "task-B")
    assert task_run.status_code == 200
    assert task_run.json["frequency"] == 10
    assert task_run.json["process"] == "FlexMeasures"
    assert task_run.json["server"] == "test"
    task_time = isodate.parse_datetime(task_run.json.get("lastrun"))
    utcnow = datetime.utcnow().replace(tzinfo=pytz.utc)
    assert task_time <= utcnow
    assert task_time >= utcnow - timedelta(minutes=2)
    assert task_run.json.get("status") == "ERROR"


def test_api_task_run_get_older_entry_then_update(client):
    task_run = get_task_run(client, "task-A")
    assert task_run.status_code == 200
    task_time = isodate.parse_datetime(task_run.json.get("lastrun"))
    utcnow = datetime.utcnow().replace(tzinfo=pytz.utc)
    assert task_time <= utcnow - timedelta(days=1)
    assert task_time >= utcnow - timedelta(days=1, minutes=1)
    assert task_run.json.get("status") == "OK"
    # update the latest run of this task (also report that it failed)
    task_update = post_task_run(client, "task-A", False)
    assert task_update.status_code == 200
    task_run = get_task_run(client, "task-A")
    task_time = isodate.parse_datetime(task_run.json.get("lastrun"))
    utcnow = datetime.utcnow().replace(tzinfo=pytz.utc)
    assert task_time <= utcnow
    assert task_time >= utcnow - timedelta(minutes=1)
    assert task_run.json.get("status") == "ERROR"



"""create_power_table_and_drop_measurement_table

Revision ID: 11b735abebe7
Revises: 1e8d27922f56
Create Date: 2018-05-29 14:51:30.331230

"""
from alembic import op
import sqlalchemy as sa
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision = "11b735abebe7"
down_revision = "1e8d27922f56"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "power",
        sa.Column("datetime", sa.DateTime(timezone=True), nullable=False),
        sa.Column("asset_id", sa.Integer(), nullable=False),
        sa.Column("value", sa.Float(), nullable=False),
        sa.ForeignKeyConstraint(["asset_id"], ["asset.id"]),
        sa.PrimaryKeyConstraint("datetime", "asset_id"),
    )
    op.drop_table("measurement")
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "measurement",
        sa.Column(
            "datetime",
            postgresql.TIMESTAMP(timezone=True),
            autoincrement=False,
            nullable=False,
        ),
        sa.Column("asset_id", sa.INTEGER(), autoincrement=False, nullable=False),
        sa.Column(
            "value",
            postgresql.DOUBLE_PRECISION(precision=53),
            autoincrement=False,
            nullable=False,
        ),
        sa.ForeignKeyConstraint(
            ["asset_id"], ["asset.id"], name="measurement_asset_id_fkey"
        ),
        sa.PrimaryKeyConstraint("datetime", "asset_id", name="measurement_pkey"),
    )
    op.drop_table("power")
    # ### end Alembic commands ###



from typing import Optional, Callable

from flask import Flask, request, jsonify, current_app
from flask_sqlalchemy import SQLAlchemy
from flask_security import Security, SQLAlchemySessionUserDatastore
from flask_login import user_logged_in
from werkzeug.exceptions import Forbidden, Unauthorized

from flexmeasures.data.models.user import User, Role, remember_login

"""
Configure auth handling.

Beware: There is a historical confusion of naming between authentication and authorization.
        Names of Responses have to be kept as they were called in original W3 protocols.
        See explanation below.
"""


# "The request requires user authentication. The response MUST include a WWW-Authenticate header field."
# So this essentially means the user needs to authenticate!
# For the historical confusion between "authorize" and "authenticate" in this status' name,
# see https://robertlathanh.com/2012/06/http-status-codes-401-unauthorized-and-403-forbidden-for-authentication-and-authorization-and-oauth/
UNAUTH_STATUS_CODE = 401
UNAUTH_ERROR_CLASS = "Unauthorized"
UNAUTH_ERROR_STATUS = (
    "UNAUTHORIZED"  # keeping the historical name intact for protocol consistency.
)
UNAUTH_MSG = (
    "You could not be properly authenticated for this content or functionality."
)

# "The server understood the request, but is refusing to fulfill it. Authorization will not help and the request SHOULD NOT be repeated."
# So this is the real authorization status!
# Preferrably to be used when the user is logged in but is not authorized for the resource.
# Advice: a not logged-in user should preferrably see a 404 NotFound.
FORBIDDEN_STATUS_CODE = 403
FORBIDDEN_ERROR_CLASS = "Forbidden"
FORBIDDEN_ERROR_STATUS = "FORBIDDEN"
FORBIDDEN_MSG = "You cannot be authorized for this content or functionality."


def configure_auth(app: Flask, db: SQLAlchemy):

    # Setup Flask-Security-Too for user authentication & authorization
    user_datastore = SQLAlchemySessionUserDatastore(db.session, User, Role)
    app.security = Security(app, user_datastore)

    # Register custom auth problem handlers.
    # Note how we are switching authorization and authentication now!
    # Flask-Security-Too seems to handle it the intendend way:
    # https://flask-security-too.readthedocs.io/en/stable/api.html#flask_security.Security.unauthn_handler
    # is defaulting to 401.
    app.security.unauthn_handler(unauthenticated_handler)
    app.register_error_handler(Unauthorized, unauthenticated_handler_e)
    app.security.unauthz_handler(unauthorized_handler)
    app.register_error_handler(Forbidden, unauthorized_handler_e)

    # add our custom handler for a user login event
    user_logged_in.connect(remember_login)


def unauthorized_handler_e(e):
    """Swallow error. Useful for classical Flask error handler registration."""
    return unauthorized_handler(None, [])


def unauthorized_handler(func: Optional[Callable], params: list):
    """
    Handler for authorization problems.
    :param func: the Flask-Security-Too decorator, if relevant, and params are its parameters.

    We support json if the request supports it.
    The ui package can also define how it wants to render HTML errors.
    """
    if func is not None:
        func(*params)
    if request.is_json:
        response = jsonify(dict(message=FORBIDDEN_MSG, status=FORBIDDEN_ERROR_STATUS))
        response.status_code = FORBIDDEN_STATUS_CODE
        return response
    elif hasattr(current_app, "unauthorized_handler_html"):
        return current_app.unauthorized_handler_html()
    else:
        return "%s:%s" % (FORBIDDEN_ERROR_CLASS, FORBIDDEN_MSG), FORBIDDEN_STATUS_CODE


def unauthenticated_handler_e(e):
    """Swallow error. Useful for classical Flask error handler registration."""
    return unauthenticated_handler([])


def unauthenticated_handler(mechanisms: list, headers: Optional[dict] = None):
    """
    Handler for authentication problems.
    :param mechanisms: a list of which authentication mechanisms were tried.
    :param headers: a dict of headers to return.
    We support json if the request supports it.
    The ui package can also define how it wants to render HTML errors.
    """
    if request.is_json:
        response = jsonify(dict(message=UNAUTH_MSG, status=UNAUTH_ERROR_STATUS))
        response.status_code = UNAUTH_STATUS_CODE
        if headers is not None:
            response.headers.update(headers)
        return response
    elif hasattr(current_app, "unauthenticated_handler_html"):
        return current_app.unauthenticated_handler_html()
    else:
        return "%s:%s" % (UNAUTH_ERROR_CLASS, UNAUTH_MSG), UNAUTH_STATUS_CODE



"""Utilities for views"""
import os
import subprocess
from typing import Tuple, List, Optional
from datetime import datetime

from flask import render_template, request, session, current_app
from bokeh.resources import CDN
from flask_security.core import current_user
from werkzeug.exceptions import BadRequest
import iso8601

from flexmeasures import __version__ as flexmeasures_version
from flexmeasures.utils import time_utils
from flexmeasures.ui import flexmeasures_ui
from flexmeasures.data.models.user import User
from flexmeasures.data.models.assets import Asset
from flexmeasures.data.models.markets import Market
from flexmeasures.data.models.weather import WeatherSensorType
from flexmeasures.data.services.resources import Resource


def render_flexmeasures_template(html_filename: str, **variables):
    """Render template and add all expected template variables, plus the ones given as **variables."""
    variables["documentation_exists"] = False
    if os.path.exists(
        "%s/static/documentation/html/index.html" % flexmeasures_ui.root_path
    ):
        variables["documentation_exists"] = True

    variables["show_queues"] = False
    if current_user.is_authenticated:
        if (
            current_user.has_role("admin")
            or current_app.config.get("FLEXMEASURES_MODE", "") == "demo"
        ):
            variables["show_queues"] = True

    variables["start_time"] = time_utils.get_default_start_time()
    if "start_time" in session:
        variables["start_time"] = session["start_time"]

    variables["end_time"] = time_utils.get_default_end_time()
    if "end_time" in session:
        variables["end_time"] = session["end_time"]

    variables["page"] = html_filename.split("/")[-1].replace(".html", "")
    if "show_datepicker" not in variables:
        variables["show_datepicker"] = variables["page"] in ("analytics", "portfolio")

    variables["contains_plots"] = False
    if any([n.endswith(("plots_div", "plots_divs")) for n in variables.keys()]):
        variables["contains_plots"] = True
        variables["bokeh_css_resources"] = CDN.render_css()
        variables["bokeh_js_resources"] = CDN.render_js()

    variables["resolution"] = session.get("resolution", "")
    variables["resolution_human"] = time_utils.freq_label_to_human_readable_label(
        session.get("resolution", "")
    )
    variables["horizon_human"] = time_utils.freq_label_to_human_readable_label(
        session.get("forecast_horizon", "")
    )

    variables["flexmeasures_version"] = flexmeasures_version

    (
        variables["git_version"],
        variables["git_commits_since"],
        variables["git_hash"],
    ) = get_git_description()
    app_start_time = current_app.config.get("START_TIME")
    variables["app_running_since"] = time_utils.naturalized_datetime_str(app_start_time)
    variables["loaded_plugins"] = ", ".join(
        f"{p_name} (v{p_version})"
        for p_name, p_version in current_app.config.get("LOADED_PLUGINS", {}).items()
    )

    variables["user_is_logged_in"] = current_user.is_authenticated
    variables[
        "user_is_admin"
    ] = current_user.is_authenticated and current_user.has_role("admin")
    variables[
        "user_is_anonymous"
    ] = current_user.is_authenticated and current_user.has_role("anonymous")
    variables["user_email"] = current_user.is_authenticated and current_user.email or ""
    variables["user_name"] = (
        current_user.is_authenticated and current_user.username or ""
    )
    variables["js_versions"] = current_app.config.get("FLEXMEASURES_JS_VERSIONS")

    return render_template(html_filename, **variables)


def clear_session():
    for skey in [
        k
        for k in session.keys()
        if k not in ("_fresh", "_id", "_user_id", "csrf_token", "fs_cc", "fs_paa")
    ]:
        current_app.logger.info(
            "Removing %s:%s from session ... " % (skey, session[skey])
        )
        del session[skey]


def set_time_range_for_session():
    """Set period (start_date, end_date and resolution) on session if they are not yet set.
    The datepicker sends times as tz-aware UTC strings.
    We re-interpret them as being in the server's timezone.
    Also set the forecast horizon, if given."""
    if "start_time" in request.values:
        session["start_time"] = time_utils.localized_datetime(
            iso8601.parse_date(request.values.get("start_time"))
        )
    elif "start_time" not in session:
        session["start_time"] = time_utils.get_default_start_time()
    else:
        if (
            session["start_time"].tzinfo is None
        ):  # session storage seems to lose tz info and becomes UTC
            session["start_time"] = time_utils.as_server_time(session["start_time"])

    if "end_time" in request.values:
        session["end_time"] = time_utils.localized_datetime(
            iso8601.parse_date(request.values.get("end_time"))
        )
    elif "end_time" not in session:
        session["end_time"] = time_utils.get_default_end_time()
    else:
        if session["end_time"].tzinfo is None:
            session["end_time"] = time_utils.as_server_time(session["end_time"])

    # Our demo server's UI should only work with the current year's data
    if current_app.config.get("FLEXMEASURES_MODE", "") == "demo":
        session["start_time"] = session["start_time"].replace(year=datetime.now().year)
        session["end_time"] = session["end_time"].replace(year=datetime.now().year)
        if session["start_time"] >= session["end_time"]:
            session["start_time"], session["end_time"] = (
                session["end_time"],
                session["start_time"],
            )

    if session["start_time"] >= session["end_time"]:
        raise BadRequest(
            "Start time %s is not after end time %s."
            % (session["start_time"], session["end_time"])
        )

    session["resolution"] = time_utils.decide_resolution(
        session["start_time"], session["end_time"]
    )

    if "forecast_horizon" in request.values:
        session["forecast_horizon"] = request.values.get("forecast_horizon")
    allowed_horizons = time_utils.forecast_horizons_for(session["resolution"])
    if (
        session.get("forecast_horizon") not in allowed_horizons
        and len(allowed_horizons) > 0
    ):
        session["forecast_horizon"] = allowed_horizons[0]


def ensure_timing_vars_are_set(
    time_window: Tuple[Optional[datetime], Optional[datetime]],
    resolution: Optional[str],
) -> Tuple[Tuple[datetime, datetime], str]:
    """
    Ensure that time window and resolution variables are set,
    even if we don't have them available ― in that case,
    get them from the session.
    """
    start = time_window[0]
    end = time_window[-1]
    if None in (start, end, resolution):
        current_app.logger.warning("Setting time range for session.")
        set_time_range_for_session()
        start_out: datetime = session["start_time"]
        end_out: datetime = session["end_time"]
        resolution_out: str = session["resolution"]
    else:
        start_out = start  # type: ignore
        end_out = end  # type: ignore
        resolution_out = resolution  # type: ignore

    return (start_out, end_out), resolution_out


def set_session_market(resource: Resource) -> Market:
    """Set session["market"] to something, based on the available markets or the request.
    Returns the selected market, or None."""
    market = resource.assets[0].market
    if market is not None:
        session["market"] = market.name
    elif "market" not in session:
        session["market"] = None
    if (
        "market" in request.args
    ):  # [GET] Set by user clicking on a link somewhere (e.g. dashboard)
        session["market"] = request.args["market"]
    if (
        "market" in request.form
    ):  # [POST] Set by user in drop-down field. This overwrites GET, as the URL remains.
        session["market"] = request.form["market"]
    return Market.query.filter(Market.name == session["market"]).one_or_none()


def set_session_sensor_type(
    accepted_sensor_types: List[WeatherSensorType],
) -> WeatherSensorType:
    """Set session["sensor_type"] to something, based on the available sensor types or the request.
    Returns the selected sensor type, or None."""

    sensor_type_name = ""
    if "sensor_type" in session:
        sensor_type_name = session["sensor_type"]
    if (
        "sensor_type" in request.args
    ):  # [GET] Set by user clicking on a link somewhere (e.g. dashboard)
        sensor_type_name = request.args["sensor_type"]
    if (
        "sensor_type" in request.form
    ):  # [POST] Set by user in drop-down field. This overwrites GET, as the URL remains.
        sensor_type_name = request.form["sensor_type"]
    requested_sensor_type = WeatherSensorType.query.filter(
        WeatherSensorType.name == sensor_type_name
    ).one_or_none()
    if (
        requested_sensor_type not in accepted_sensor_types
        and len(accepted_sensor_types) > 0
    ):
        sensor_type = accepted_sensor_types[0]
        session["sensor_type"] = sensor_type.name
        return sensor_type
    elif len(accepted_sensor_types) == 0:
        session["sensor_type"] = None
    else:
        session["sensor_type"] = requested_sensor_type.name
        return requested_sensor_type


def set_session_resource(
    assets: List[Asset], groups_with_assets: List[str]
) -> Optional[Resource]:
    """
    Set session["resource"] to something, based on the available asset groups or the request.

    Returns the selected resource instance, or None.
    """
    if (
        "resource" in request.args
    ):  # [GET] Set by user clicking on a link somewhere (e.g. dashboard)
        session["resource"] = request.args["resource"]
    if (
        "resource" in request.form
    ):  # [POST] Set by user in drop-down field. This overwrites GET, as the URL remains.
        session["resource"] = request.form["resource"]

    if "resource" not in session:  # set some default, if possible
        if len(groups_with_assets) > 0:
            session["resource"] = groups_with_assets[0]
        elif len(assets) > 0:
            session["resource"] = assets[0].name
        else:
            return None

    return Resource(session["resource"])


def set_individual_traces_for_session():
    """
    Set session["showing_individual_traces_for"] to a value ("none", "power", "schedules").
    """
    var_name = "showing_individual_traces_for"
    if var_name not in session:
        session[var_name] = "none"  # default setting: we show traces aggregated
    if var_name in request.values and request.values[var_name] in (
        "none",
        "power",
        "schedules",
    ):
        session[var_name] = request.values[var_name]


def get_git_description() -> Tuple[str, int, str]:
    """
    Get information about the SCM (git) state if possible (if a .git directory exists).

    Returns the latest git version (tag) as a string, the number of commits since then as an int and the
    current commit hash as string.
    """

    def _minimal_ext_cmd(cmd: list):
        # construct minimal environment
        env = {}
        for k in ["SYSTEMROOT", "PATH"]:
            v = os.environ.get(k)
            if v is not None:
                env[k] = v
        # LANGUAGE is used on win32
        env["LANGUAGE"] = "C"
        env["LANG"] = "C"
        env["LC_ALL"] = "C"
        return subprocess.Popen(cmd, stdout=subprocess.PIPE, env=env).communicate()[0]

    version = "Unknown"
    commits_since = 0
    sha = "Unknown"

    path_to_flexmeasures_root = os.path.join(
        os.path.dirname(__file__), "..", "..", ".."
    )
    if os.path.exists(os.path.join(path_to_flexmeasures_root, ".git")):
        commands = ["git", "describe", "--always", "--long"]
        try:
            git_output = _minimal_ext_cmd(commands)
            components = git_output.strip().decode("ascii").split("-")
            if not (len(components) == 1 and components[0] == ""):
                sha = components.pop()
                if len(components) > 0:
                    commits_since = int(components.pop())
                    version = "-".join(components)
        except OSError as ose:
            current_app.logger.warning("Problem when reading git describe: %s" % ose)

    return version, commits_since, sha


def asset_icon_name(asset_type_name: str) -> str:
    """Icon name for this asset type.

    This can be used for UI html templates made with Jinja.
    ui.__init__ makes this function available as the filter "asset_icon".

    For example:
        <i class={{ asset_type.name | asset_icon }}></i>
    becomes (for a battery):
        <i class="icon-battery"></i>
    """
    # power asset exceptions
    if "evse" in asset_type_name.lower():
        return "icon-charging_station"
    # weather exceptions
    if asset_type_name == "radiation":
        return "wi wi-horizon-alt"
    elif asset_type_name == "temperature":
        return "wi wi-thermometer"
    elif asset_type_name == "wind_direction":
        return "wi wi-wind-direction"
    elif asset_type_name == "wind_speed":
        return "wi wi-strong-wind"
    # aggregation exceptions
    elif asset_type_name == "renewables":
        return "icon-wind"
    return f"icon-{asset_type_name}"


def username(user_id) -> str:
    user = User.query.get(user_id)
    if user is None:
        current_app.logger.warning(f"Could not find user with id {user_id}")
        return ""
    else:
        return user.username



from datetime import timedelta

import pytest
from flask import url_for
from iso8601 import parse_date

from flexmeasures.api.common.schemas.sensors import SensorField
from flexmeasures.api.tests.utils import get_auth_token
from flexmeasures.api.v2_0.tests.utils import (
    message_for_post_price_data,
    verify_sensor_data_in_db,
)


@pytest.mark.parametrize(
    "post_message",
    [
        message_for_post_price_data(market_id=7),
        message_for_post_price_data(market_id=1, prior_instead_of_horizon=True),
    ],
)
def test_post_price_data_2_0(
    fresh_db,
    setup_roles_users_fresh_db,
    setup_markets_fresh_db,
    clean_redis,
    app,
    post_message,
):
    """
    Try to post price data as a logged-in test user with the Supplier role, which should succeed.
    """
    db = fresh_db
    # call with client whose context ends, so that we can test for,
    # after-effects in the database after teardown committed.
    with app.test_client() as client:
        # post price data
        auth_token = get_auth_token(client, "test_supplier@seita.nl", "testtest")
        post_price_data_response = client.post(
            url_for("flexmeasures_api_v2_0.post_price_data"),
            json=post_message,
            headers={"Authorization": auth_token},
        )
        print("Server responded with:\n%s" % post_price_data_response.json)
        assert post_price_data_response.status_code == 200
        assert post_price_data_response.json["type"] == "PostPriceDataResponse"

    verify_sensor_data_in_db(
        post_message, post_message["values"], db, entity_type="market", fm_scheme="fm1"
    )

    # look for Forecasting jobs in queue
    assert (
        len(app.queues["forecasting"]) == 2
    )  # only one market is affected, but two horizons
    horizons = [timedelta(hours=24), timedelta(hours=48)]
    jobs = sorted(app.queues["forecasting"].jobs, key=lambda x: x.kwargs["horizon"])
    market = SensorField("market", fm_scheme="fm1").deserialize(post_message["market"])
    for job, horizon in zip(jobs, horizons):
        assert job.kwargs["horizon"] == horizon
        assert job.kwargs["start"] == parse_date(post_message["start"]) + horizon
        assert job.kwargs["timed_value_type"] == "Price"
        assert job.kwargs["asset_id"] == market.id



"""add_location_columns_to_weather_sensor_table

Revision ID: 9c7fc8e46f1e
Revises: 2c9a32614784
Create Date: 2018-07-27 12:38:27.006000

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = "9c7fc8e46f1e"
down_revision = "2c9a32614784"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    # We need to add the nullable constraint after we set default values for any existing sensors.
    # These only matter for our current development, not when you set up a new database!
    # This default location is Jeju island.
    op.add_column("weather_sensor", sa.Column("latitude", sa.Float()))
    op.execute("UPDATE weather_sensor SET latitude = 33.4843866")
    op.alter_column("weather_sensor", "latitude", nullable=False)
    op.add_column("weather_sensor", sa.Column("longitude", sa.Float()))
    op.execute("UPDATE weather_sensor SET longitude = 126.477859")
    op.alter_column("weather_sensor", "longitude", nullable=False)
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("weather_sensor", "latitude")
    op.drop_column("weather_sensor", "longitude")
    # ### end Alembic commands ###



"""mix in timely beliefs sensor with asset, market and weather sensor; introduce knowledge horizons

Revision ID: 22ce09690d23
Revises: 564e8df4e3a9
Create Date: 2021-01-31 14:31:16.370110

"""
from alembic import op
import json
import sqlalchemy as sa
from timely_beliefs.sensors.func_store import knowledge_horizons


# revision identifiers, used by Alembic.
revision = "22ce09690d23"
down_revision = "564e8df4e3a9"
branch_labels = None
depends_on = None

# set default parameters for the two default knowledge horizon functions
ex_ante_default_par = {knowledge_horizons.ex_ante.__code__.co_varnames[0]: "PT0H"}
ex_post_default_par = {knowledge_horizons.ex_post.__code__.co_varnames[1]: "PT0H"}


def upgrade():

    # Mix in timely_beliefs.Sensor with flexmeasures.Asset
    op.add_column(
        "asset",
        sa.Column(
            "knowledge_horizon_fnc",
            sa.String(length=80),
            nullable=True,
            default=knowledge_horizons.ex_post.__name__,
        ),
    )
    op.execute(
        f"update asset set knowledge_horizon_fnc = '{knowledge_horizons.ex_post.__name__}';"
    )  # default assumption that power measurements are known right after the fact
    op.alter_column("asset", "knowledge_horizon_fnc", nullable=False)

    op.add_column(
        "asset",
        sa.Column(
            "knowledge_horizon_par",
            sa.JSON(),
            nullable=True,
            default={knowledge_horizons.ex_post.__code__.co_varnames[1]: "PT0H"},
        ),
    )
    op.execute(
        f"""update asset set knowledge_horizon_par = '{json.dumps(ex_post_default_par)}';"""
    )
    op.alter_column("asset", "knowledge_horizon_par", nullable=False)

    op.add_column("asset", sa.Column("timezone", sa.String(length=80), nullable=True))
    op.execute("update asset set timezone = 'Asia/Seoul';")
    op.alter_column("asset", "timezone", nullable=False)

    # Mix in timely_beliefs.Sensor with flexmeasures.Market
    op.add_column(
        "market",
        sa.Column(
            "knowledge_horizon_fnc",
            sa.String(length=80),
            nullable=True,
            default=knowledge_horizons.ex_ante.__name__,
        ),
    )
    op.execute(
        f"update market set knowledge_horizon_fnc = '{knowledge_horizons.ex_ante.__name__}';"
    )  # default assumption that prices are known before a transaction
    op.execute(
        f"update market set knowledge_horizon_fnc = '{knowledge_horizons.at_date.__name__}' where name in ('kepco_cs_fast', 'kepco_cs_slow', 'kepco_cs_smart');"
    )
    op.alter_column("market", "knowledge_horizon_fnc", nullable=False)

    op.add_column(
        "market",
        sa.Column(
            "knowledge_horizon_par",
            sa.JSON(),
            nullable=True,
            default=ex_ante_default_par,
        ),
    )
    op.execute(
        f"""update market set knowledge_horizon_par = '{json.dumps(ex_ante_default_par)}';"""
    )
    op.execute(
        """update market set knowledge_horizon_par = '{"knowledge_time": "2014-12-31 00:00:00+00:00"}' where name in ('kepco_cs_fast', 'kepco_cs_slow', 'kepco_cs_smart');"""
    )  # tariff publication date (unofficial)
    op.alter_column("market", "knowledge_horizon_par", nullable=False)
    op.execute(
        "update price set horizon = interval '0 hours' from market where market_id = market.id and market.name in ('kepco_cs_fast', 'kepco_cs_slow', 'kepco_cs_smart');"
    )  # 0 hours after fixed knowledge time (i.e. at publication date)

    op.add_column("market", sa.Column("timezone", sa.String(length=80), nullable=True))
    op.execute("update market set timezone = 'UTC';")
    op.execute("update market set timezone = 'Europe/Paris' where name='epex_da';")
    op.execute("update market set timezone = 'Asia/Seoul' where unit='KRW/kWh';")
    op.alter_column("market", "timezone", nullable=False)

    # Mix in timely_beliefs.Sensor with flexmeasures.WeatherSensor
    op.add_column(
        "weather_sensor",
        sa.Column(
            "knowledge_horizon_fnc",
            sa.String(length=80),
            nullable=True,
            default=knowledge_horizons.ex_post.__name__,
        ),
    )
    op.execute(
        f"update weather_sensor set knowledge_horizon_fnc = '{knowledge_horizons.ex_post.__name__}';"
    )  # default assumption that weather measurements are known right after the fact
    op.alter_column("weather_sensor", "knowledge_horizon_fnc", nullable=False)

    op.add_column(
        "weather_sensor",
        sa.Column(
            "knowledge_horizon_par",
            sa.JSON(),
            nullable=True,
            default={knowledge_horizons.ex_post.__code__.co_varnames[1]: "PT0H"},
        ),
    )
    op.execute(
        f"""update weather_sensor set knowledge_horizon_par = '{json.dumps(ex_post_default_par)}';"""
    )
    op.alter_column("weather_sensor", "knowledge_horizon_par", nullable=False)

    op.add_column(
        "weather_sensor", sa.Column("timezone", sa.String(length=80), nullable=True)
    )
    op.execute("update weather_sensor set timezone = 'Asia/Seoul';")
    op.alter_column("weather_sensor", "timezone", nullable=False)

    # todo: execute after adding relevant tests and updating our strategy to create forecasting jobs when new information arrives
    # op.execute(
    #     f"update market set knowledge_horizon_fnc = '{knowledge_horizons.x_days_ago_at_y_oclock.__name__}' where name in ('epex_da', 'kpx_da');"
    # )
    # op.execute(
    #     """update market set knowledge_horizon_par = '{"x": 1, "y": 12, "z": "Europe/Paris"}' where name='epex_da';"""
    # )  # gate closure at 12:00 on the preceding day, with expected price publication at 12.42 and 12.55 (from EPEX Spot Day-Ahead Multi-Regional Coupling, https://www.epexspot.com/en/downloads#rules-fees-processes )
    # op.execute(
    #     """update market set knowledge_horizon_par = '{"x": 1, "y": 10, "z": "Asia/Seoul"}' where name='kpx_da';"""
    # )  # gate closure at 10.00 on the preceding day, with expected price publication at 15.00 (from KPX Power Market Operation, https://www.slideshare.net/sjchung0/power-market-operation )
    # todo: add statement to update the horizon for prices on these markets, in accordance with their new knowledge horizon function


def downgrade():
    # Drop mixed in columns
    op.drop_column("asset", "timezone")
    op.drop_column("asset", "knowledge_horizon_par")
    op.drop_column("asset", "knowledge_horizon_fnc")
    op.drop_column("market", "timezone")
    op.drop_column("market", "knowledge_horizon_par")
    op.drop_column("market", "knowledge_horizon_fnc")
    op.drop_column("weather_sensor", "timezone")
    op.drop_column("weather_sensor", "knowledge_horizon_par")
    op.drop_column("weather_sensor", "knowledge_horizon_fnc")
    op.execute(
        "update price set horizon = ((datetime + market.event_resolution) - '2014-12-31 00:00:00+00:00') from market where market_id = market.id and name in ('kepco_cs_fast', 'kepco_cs_slow', 'kepco_cs_smart');"
    )  # rolling horizon before end of event (i.e. at publication date)



from typing import List, Optional, Union, Tuple, Dict
from datetime import datetime, timedelta
import io
import csv
import json
import math

import pandas as pd
from flask import session, current_app, make_response
from flask_security import roles_accepted
from bokeh.plotting import Figure
from bokeh.embed import components
from bokeh.util.string import encode_utf8
from bokeh.models import Range1d
from pandas.tseries.frequencies import to_offset

from flexmeasures.data.models.markets import Market
from flexmeasures.data.models.weather import WeatherSensor
from flexmeasures.data.services.resources import (
    get_assets,
    get_asset_group_queries,
    get_markets,
    get_sensor_types,
    Resource,
)
from flexmeasures.data.queries.analytics import (
    get_power_data,
    get_prices_data,
    get_weather_data,
    get_revenues_costs_data,
)
from flexmeasures.utils import time_utils
from flexmeasures.utils.flexmeasures_inflection import humanize
from flexmeasures.ui.utils.view_utils import (
    render_flexmeasures_template,
    set_session_resource,
    set_session_market,
    set_session_sensor_type,
    set_individual_traces_for_session,
    set_time_range_for_session,
    ensure_timing_vars_are_set,
)
from flexmeasures.ui.utils.plotting_utils import create_graph, separate_legend
from flexmeasures.ui.views import flexmeasures_ui


@flexmeasures_ui.route("/analytics", methods=["GET", "POST"])
@roles_accepted("admin", "Prosumer")
def analytics_view():
    """Analytics view. Here, four plots (consumption/generation, weather, prices and a profit/loss calculation)
    and a table of metrics data are prepared. This view allows to select a resource name, from which a
    `models.Resource` object can be made. The resource name is kept in the session.
    Based on the resource, plots and table are labelled appropriately.
    """
    set_time_range_for_session()
    markets = get_markets()
    assets = get_assets(order_by_asset_attribute="display_name", order_direction="asc")
    asset_groups = get_asset_group_queries(
        custom_additional_groups=["renewables", "EVSE", "location"]
    )
    asset_group_names: List[str] = [
        group for group in asset_groups if asset_groups[group].count() > 0
    ]
    selected_resource = set_session_resource(assets, asset_group_names)
    if selected_resource is None:
        raise Exception(
            "No assets exist yet, so the analytics view will not work. Please add an asset!"
        )

    selected_market = set_session_market(selected_resource)
    sensor_types = get_sensor_types(selected_resource)
    session_asset_types = selected_resource.unique_asset_types
    selected_sensor_type = set_session_sensor_type(sensor_types)
    set_individual_traces_for_session()
    view_shows_individual_traces = (
        session["showing_individual_traces_for"] in ("power", "schedules")
        and selected_resource.is_eligible_for_comparing_individual_traces()
    )

    query_window, resolution = ensure_timing_vars_are_set((None, None), None)

    # This is useful information - we might want to adapt the sign of the data and labels.
    showing_pure_consumption_data = all(
        [a.is_pure_consumer for a in selected_resource.assets]
    )
    showing_pure_production_data = all(
        [a.is_pure_producer for a in selected_resource.assets]
    )
    # Only show production positive if all assets are producers
    show_consumption_as_positive = False if showing_pure_production_data else True

    # ---- Get data

    data, metrics, weather_type, selected_weather_sensor = get_data_and_metrics(
        query_window,
        resolution,
        show_consumption_as_positive,
        session["showing_individual_traces_for"]
        if view_shows_individual_traces
        else "none",
        selected_resource,
        selected_market,
        selected_sensor_type,
        selected_resource.assets,
    )

    # TODO: get rid of these hacks, which we use because we mock the current year's data from 2015 data in demo mode
    # Our demo server uses 2015 data as if it's the current year's data. Here we mask future beliefs.
    if current_app.config.get("FLEXMEASURES_MODE", "") == "demo":
        data = filter_for_past_data(data)
        data = filter_forecasts_for_limited_time_window(data)

    # ---- Making figures

    # Set shared x range
    shared_x_range = Range1d(start=query_window[0], end=query_window[1])
    shared_x_range2 = Range1d(
        start=query_window[0], end=query_window[1]
    )  # only needed if we draw two legends (if individual traces are on)

    tools = ["box_zoom", "reset", "save"]
    power_fig = make_power_figure(
        selected_resource.display_name,
        data["power"],
        data["power_forecast"],
        data["power_schedule"],
        show_consumption_as_positive,
        shared_x_range,
        tools=tools,
    )
    rev_cost_fig = make_revenues_costs_figure(
        selected_resource.display_name,
        data["rev_cost"],
        data["rev_cost_forecast"],
        show_consumption_as_positive,
        shared_x_range,
        selected_market,
        tools=tools,
    )
    # the bottom plots need a separate x axis if they get their own legend (Bokeh complains otherwise)
    # this means in that in that corner case zooming will not work across all four plots
    prices_fig = make_prices_figure(
        data["prices"],
        data["prices_forecast"],
        shared_x_range2 if view_shows_individual_traces else shared_x_range,
        selected_market,
        tools=tools,
    )
    weather_fig = make_weather_figure(
        selected_resource,
        data["weather"],
        data["weather_forecast"],
        shared_x_range2 if view_shows_individual_traces else shared_x_range,
        selected_weather_sensor,
        tools=tools,
    )

    # Separate a single legend and remove the others.
    # In case of individual traces, we need two legends.
    top_legend_fig = separate_legend(power_fig, orientation="horizontal")
    top_legend_script, top_legend_div = components(top_legend_fig)
    rev_cost_fig.renderers.remove(rev_cost_fig.legend[0])
    if view_shows_individual_traces:
        bottom_legend_fig = separate_legend(weather_fig, orientation="horizontal")
        prices_fig.renderers.remove(prices_fig.legend[0])
        bottom_legend_script, bottom_legend_div = components(bottom_legend_fig)
    else:
        prices_fig.renderers.remove(prices_fig.legend[0])
        weather_fig.renderers.remove(weather_fig.legend[0])
        bottom_legend_fig = bottom_legend_script = bottom_legend_div = None

    analytics_plots_script, analytics_plots_divs = components(
        (power_fig, rev_cost_fig, prices_fig, weather_fig)
    )

    return render_flexmeasures_template(
        "views/analytics.html",
        top_legend_height=top_legend_fig.plot_height,
        top_legend_script=top_legend_script,
        top_legend_div=top_legend_div,
        bottom_legend_height=0
        if bottom_legend_fig is None
        else bottom_legend_fig.plot_height,
        bottom_legend_script=bottom_legend_script,
        bottom_legend_div=bottom_legend_div,
        analytics_plots_divs=[encode_utf8(div) for div in analytics_plots_divs],
        analytics_plots_script=analytics_plots_script,
        metrics=metrics,
        markets=markets,
        sensor_types=sensor_types,
        assets=assets,
        asset_group_names=asset_group_names,
        selected_market=selected_market,
        selected_resource=selected_resource,
        selected_sensor_type=selected_sensor_type,
        selected_sensor=selected_weather_sensor,
        asset_types=session_asset_types,
        showing_pure_consumption_data=showing_pure_consumption_data,
        showing_pure_production_data=showing_pure_production_data,
        show_consumption_as_positive=show_consumption_as_positive,
        showing_individual_traces_for=session["showing_individual_traces_for"],
        offer_showing_individual_traces=selected_resource.is_eligible_for_comparing_individual_traces(),
        forecast_horizons=time_utils.forecast_horizons_for(session["resolution"]),
        active_forecast_horizon=session["forecast_horizon"],
    )


@flexmeasures_ui.route("/analytics_data/<content>/<content_type>", methods=["GET"])
@roles_accepted("admin", "Prosumer")
def analytics_data_view(content, content_type):
    """Analytics view as above, but here we only download data.
    Content can be either source or metrics.
    Content-type can be either CSV or JSON.
    """
    # if current_app.config.get("FLEXMEASURES_MODE", "") != "play":
    #    raise NotImplementedError("The analytics data download only works in play mode.")
    if content not in ("source", "metrics"):
        if content is None:
            content = "data"
        else:
            raise NotImplementedError("content can either be source or metrics.")
    if content_type not in ("csv", "json"):
        if content_type is None:
            content_type = "csv"
        else:
            raise NotImplementedError("content_type can either be csv or json.")

    time_utils.set_time_range_for_session()

    # Maybe move some of this stuff into get_data_and_metrics
    assets = get_assets(order_by_asset_attribute="display_name", order_direction="asc")
    asset_groups = get_asset_group_queries(
        custom_additional_groups=["renewables", "EVSE", "location"]
    )
    asset_group_names: List[str] = [
        group for group in asset_groups if asset_groups[group].count() > 0
    ]
    selected_resource = set_session_resource(assets, asset_group_names)
    selected_market = set_session_market(selected_resource)
    sensor_types = get_sensor_types(selected_resource)
    selected_sensor_type = set_session_sensor_type(sensor_types)

    # This is useful information - we might want to adapt the sign of the data and labels.
    showing_pure_consumption_data = all(
        [a.is_pure_consumer for a in selected_resource.assets]
    )
    showing_pure_production_data = all(
        [a.is_pure_producer for a in selected_resource.assets]
    )
    # Only show production positive if all assets are producers
    show_consumption_as_positive = False if showing_pure_production_data else True

    # Getting data and calculating metrics for them
    query_window, resolution = ensure_timing_vars_are_set((None, None), None)
    data, metrics, weather_type, selected_weather_sensor = get_data_and_metrics(
        query_window,
        resolution,
        show_consumption_as_positive,
        "none",
        selected_resource,
        selected_market,
        selected_sensor_type,
        selected_resource.assets,
    )

    horizon = session["forecast_horizon"]
    rev_cost_header = (
        "costs/revenues" if show_consumption_as_positive else "revenues/costs"
    )
    if showing_pure_consumption_data:
        rev_cost_header = "costs"
    elif showing_pure_production_data:
        rev_cost_header = "revenues"
    source_headers = [
        "time",
        "power_data_label",
        "power",
        "power_forecast_label",
        f"power_forecast_{horizon}",
        f"{weather_type}_label",
        f"{weather_type}",
        f"{weather_type}_forecast_label",
        f"{weather_type}_forecast_{horizon}",
        "price_label",
        f"price_on_{selected_market.name}",
        "price_forecast_label",
        f"price_forecast_{horizon}",
        f"{rev_cost_header}_label",
        rev_cost_header,
        f"{rev_cost_header}_forecast_label",
        f"{rev_cost_header}_forecast_{horizon}",
    ]
    source_units = [
        "",
        "",
        "MW",
        "",
        "MW",
        "",
        selected_weather_sensor.unit,
        "",
        selected_weather_sensor.unit,
        "",
        selected_market.price_unit,
        "",
        selected_market.price_unit,
        "",
        selected_market.price_unit[:3],
        "",
        selected_market.price_unit[:3],
    ]
    filename_prefix = "%s_analytics" % selected_resource.name
    if content_type == "csv":
        str_io = io.StringIO()
        writer = csv.writer(str_io, dialect="excel")
        if content == "metrics":
            filename = "%s_metrics.csv" % filename_prefix
            writer.writerow(metrics.keys())
            writer.writerow(metrics.values())
        else:
            filename = "%s_source.csv" % filename_prefix
            writer.writerow(source_headers)
            writer.writerow(source_units)
            for dt in data["rev_cost"].index:
                row = [
                    dt,
                    data["power"].loc[dt].label
                    if "label" in data["power"].columns
                    else "Aggregated power data",
                    data["power"].loc[dt]["event_value"],
                    data["power_forecast"].loc[dt].label
                    if "label" in data["power_forecast"].columns
                    else "Aggregated power forecast data",
                    data["power_forecast"].loc[dt]["event_value"],
                    data["weather"].loc[dt].label
                    if "label" in data["weather"].columns
                    else f"Aggregated {weather_type} data",
                    data["weather"].loc[dt]["event_value"],
                    data["weather_forecast"].loc[dt].label
                    if "label" in data["weather_forecast"].columns
                    else f"Aggregated {weather_type} forecast data",
                    data["weather_forecast"].loc[dt]["event_value"],
                    data["prices"].loc[dt].label
                    if "label" in data["prices"].columns
                    else "Aggregated power data",
                    data["prices"].loc[dt]["event_value"],
                    data["prices_forecast"].loc[dt].label
                    if "label" in data["prices_forecast"].columns
                    else "Aggregated power data",
                    data["prices_forecast"].loc[dt]["event_value"],
                    data["rev_cost"].loc[dt].label
                    if "label" in data["rev_cost"].columns
                    else f"Aggregated {rev_cost_header} data",
                    data["rev_cost"].loc[dt]["event_value"],
                    data["rev_cost_forecast"].loc[dt].label
                    if "label" in data["rev_cost_forecast"].columns
                    else f"Aggregated {rev_cost_header} forecast data",
                    data["rev_cost_forecast"].loc[dt]["event_value"],
                ]
                writer.writerow(row)

        response = make_response(str_io.getvalue())
        response.headers["Content-Disposition"] = "attachment; filename=%s" % filename
        response.headers["Content-type"] = "text/csv"
    else:
        if content == "metrics":
            filename = "%s_metrics.json" % filename_prefix
            response = make_response(json.dumps(metrics))
        else:
            # Not quite done yet. I don't like how we treat forecasts in here yet. Not sure how to mention units.
            filename = "%s_source.json" % filename_prefix
            json_strings = []
            for key in data:
                json_strings.append(
                    f"\"{key}\":{data[key].to_json(orient='index', date_format='iso')}"
                )
            response = make_response("{%s}" % ",".join(json_strings))
        response.headers["Content-Disposition"] = "attachment; filename=%s" % filename
        response.headers["Content-type"] = "application/json"
    return response


def get_data_and_metrics(
    query_window: Tuple[datetime, datetime],
    resolution: str,
    show_consumption_as_positive: bool,
    showing_individual_traces_for: str,
    selected_resource: Resource,
    selected_market,
    selected_sensor_type,
    assets,
) -> Tuple[Dict[str, pd.DataFrame], Dict[str, float], str, WeatherSensor]:
    """Getting data and calculating metrics for them"""
    data: Dict[str, pd.DataFrame] = dict()
    forecast_horizon = pd.to_timedelta(session["forecast_horizon"])
    metrics: dict = dict()
    (
        data["power"],
        data["power_forecast"],
        data["power_schedule"],
        metrics,
    ) = get_power_data(
        selected_resource,
        show_consumption_as_positive,
        showing_individual_traces_for,
        metrics,
        query_window,
        resolution,
        forecast_horizon,
    )
    data["prices"], data["prices_forecast"], metrics = get_prices_data(
        metrics,
        selected_market,
        query_window,
        resolution,
        forecast_horizon,
    )
    (
        data["weather"],
        data["weather_forecast"],
        weather_type,
        selected_sensor,
        metrics,
    ) = get_weather_data(
        assets,
        metrics,
        selected_sensor_type,
        query_window,
        resolution,
        forecast_horizon,
    )
    # TODO: get rid of this hack, which we use because we mock forecast intervals in demo mode
    if current_app.config.get("FLEXMEASURES_MODE", "") == "demo":
        # In each case below, the error increases with the horizon towards a certain percentage of the point forecast
        horizon = pd.to_timedelta(to_offset(session.get("forecast_horizon", "1d")))
        decay_factor = 1 - math.exp(-horizon / timedelta(hours=6))

        # Heuristic power confidence interval
        error_margin = 0.1 * decay_factor
        data["power_forecast"]["yhat_upper"] = data["power_forecast"]["event_value"] * (
            1 + error_margin
        )
        data["power_forecast"]["yhat_lower"] = data["power_forecast"]["event_value"] * (
            1 - error_margin
        )

        # Heuristic price confidence interval
        error_margin_upper = 0.6 * decay_factor
        error_margin_lower = 0.3 * decay_factor
        data["prices_forecast"]["yhat_upper"] = data["prices_forecast"][
            "event_value"
        ] * (1 + error_margin_upper)
        data["prices_forecast"]["yhat_lower"] = data["prices_forecast"][
            "event_value"
        ] * (1 - error_margin_lower)

        # Heuristic weather confidence interval
        if weather_type == "temperature":
            error_margin_upper = 0.7 * decay_factor
            error_margin_lower = error_margin_upper
        elif weather_type == "wind_speed":
            error_margin_upper = 1.5 * decay_factor
            error_margin_lower = 0.8 * decay_factor
        elif weather_type == "radiation":
            error_margin_upper = 1.8 * decay_factor
            error_margin_lower = 0.5 * decay_factor
        if data["weather_forecast"].empty:
            data["weather_forecast"] = data["weather"].copy()
            data["weather_forecast"]["event_value"] *= 1.1
        data["weather_forecast"]["yhat_upper"] = data["weather_forecast"][
            "event_value"
        ] * (1 + error_margin_upper)
        data["weather_forecast"]["yhat_lower"] = data["weather_forecast"][
            "event_value"
        ] * (1 - error_margin_lower)

    unit_factor = revenue_unit_factor("MWh", selected_market.unit)
    data["rev_cost"], data["rev_cost_forecast"], metrics = get_revenues_costs_data(
        data["power"],
        data["prices"],
        data["power_forecast"],
        data["prices_forecast"],
        metrics,
        unit_factor,
        resolution,
        showing_individual_traces_for in ("power", "schedules"),
    )
    return data, metrics, weather_type, selected_sensor


def filter_for_past_data(data):
    """ Make sure we only show past data, useful for demo mode """
    most_recent_quarter = time_utils.get_most_recent_quarter()

    if not data["power"].empty:
        data["power"] = data["power"].loc[
            data["power"].index.get_level_values("event_start") < most_recent_quarter
        ]
    if not data["prices"].empty:
        data["prices"] = data["prices"].loc[
            data["prices"].index < most_recent_quarter + timedelta(hours=24)
        ]  # keep tomorrow's prices
    if not data["weather"].empty:
        data["weather"] = data["weather"].loc[
            data["weather"].index < most_recent_quarter
        ]
    if not data["rev_cost"].empty:
        data["rev_cost"] = data["rev_cost"].loc[
            data["rev_cost"].index.get_level_values("event_start") < most_recent_quarter
        ]
    return data


def filter_forecasts_for_limited_time_window(data):
    """ Show forecasts only up to a limited horizon """
    most_recent_quarter = time_utils.get_most_recent_quarter()
    horizon_days = 10  # keep a 10 day forecast
    max_forecast_datetime = most_recent_quarter + timedelta(hours=horizon_days * 24)
    if not data["power_forecast"].empty:
        data["power_forecast"] = data["power_forecast"].loc[
            data["power_forecast"].index < max_forecast_datetime
        ]
    if not data["prices_forecast"].empty:
        data["prices_forecast"] = data["prices_forecast"].loc[
            data["prices_forecast"].index < max_forecast_datetime
        ]
    if not data["weather_forecast"].empty:
        data["weather_forecast"] = data["weather_forecast"].loc[
            data["weather_forecast"].index < max_forecast_datetime
        ]
    if not data["rev_cost_forecast"].empty:
        data["rev_cost_forecast"] = data["rev_cost_forecast"].loc[
            data["rev_cost_forecast"].index < max_forecast_datetime
        ]
    return data


def make_power_figure(
    resource_display_name: str,
    data: pd.DataFrame,
    forecast_data: Optional[pd.DataFrame],
    schedule_data: Optional[pd.DataFrame],
    show_consumption_as_positive: bool,
    shared_x_range: Range1d,
    tools: List[str] = None,
    sizing_mode="scale_width",
) -> Figure:
    """Make a bokeh figure for power consumption or generation"""
    if show_consumption_as_positive:
        title = "Electricity consumption of %s" % resource_display_name
    else:
        title = "Electricity production from %s" % resource_display_name
    if data.empty:
        title = title.replace("Electricity", "Prognosed")

    return create_graph(
        data,
        unit="MW",
        legend_location="top_right",
        legend_labels=("Actual", "Forecast", None)
        if schedule_data is None or schedule_data["event_value"].isnull().all()
        else ("Actual", "Forecast", "Schedule"),
        forecasts=forecast_data,
        schedules=schedule_data,
        title=title,
        x_range=shared_x_range,
        x_label="Time (resolution of %s)"
        % determine_resolution(data, forecast_data, schedule_data),
        y_label="Power (in MW)",
        show_y_floats=True,
        tools=tools,
        sizing_mode=sizing_mode,
    )


def make_prices_figure(
    data: pd.DataFrame,
    forecast_data: Union[None, pd.DataFrame],
    shared_x_range: Range1d,
    selected_market: Market,
    tools: List[str] = None,
    sizing_mode="scale_width",
) -> Figure:
    """Make a bokeh figure for price data"""
    return create_graph(
        data,
        unit=selected_market.unit,
        legend_location="top_right",
        forecasts=forecast_data,
        title=f"Prices for {selected_market.display_name}",
        x_range=shared_x_range,
        x_label="Time (resolution of %s)" % determine_resolution(data, forecast_data),
        y_label="Price (in %s)" % selected_market.unit,
        show_y_floats=True,
        tools=tools,
        sizing_mode=sizing_mode,
    )


def make_weather_figure(
    selected_resource: Resource,
    data: pd.DataFrame,
    forecast_data: Union[None, pd.DataFrame],
    shared_x_range: Range1d,
    weather_sensor: WeatherSensor,
    tools: List[str] = None,
    sizing_mode="scale_width",
) -> Figure:
    """Make a bokeh figure for weather data"""
    # Todo: plot average temperature/total_radiation/wind_speed for asset groups, and update title accordingly
    if weather_sensor is None:
        return create_graph(
            pd.DataFrame(columns=["event_value"]),
            title="Weather plot (no relevant weather sensor found)",
        )
    unit = weather_sensor.unit
    weather_axis_label = "%s (in %s)" % (
        humanize(weather_sensor.sensor_type.display_name),
        unit,
    )

    if selected_resource.is_unique_asset:
        title = "%s at %s" % (
            humanize(weather_sensor.sensor_type.display_name),
            selected_resource.display_name,
        )
    else:
        title = "%s" % humanize(weather_sensor.sensor_type.display_name)
    return create_graph(
        data,
        unit=unit,
        forecasts=forecast_data,
        title=title,
        x_range=shared_x_range,
        x_label="Time (resolution of %s)" % determine_resolution(data, forecast_data),
        y_label=weather_axis_label,
        legend_location="top_right",
        show_y_floats=True,
        tools=tools,
        sizing_mode=sizing_mode,
    )


def make_revenues_costs_figure(
    resource_display_name: str,
    data: pd.DataFrame,
    forecast_data: pd.DataFrame,
    show_consumption_as_positive: bool,
    shared_x_range: Range1d,
    selected_market: Market,
    tools: List[str] = None,
    sizing_mode="scale_width",
) -> Figure:
    """Make a bokeh figure for revenues / costs data"""
    if show_consumption_as_positive:
        rev_cost_str = "Costs"
    else:
        rev_cost_str = "Revenues"

    return create_graph(
        data,
        unit=selected_market.unit[
            :3
        ],  # First three letters of a price unit give the currency (ISO 4217)
        legend_location="top_right",
        forecasts=forecast_data,
        title=f"{rev_cost_str} for {resource_display_name} (on {selected_market.display_name})",
        x_range=shared_x_range,
        x_label="Time (resolution of %s)" % determine_resolution(data, forecast_data),
        y_label="%s (in %s)" % (rev_cost_str, selected_market.unit[:3]),
        show_y_floats=True,
        tools=tools,
        sizing_mode=sizing_mode,
    )


def revenue_unit_factor(quantity_unit: str, price_unit: str) -> float:
    market_quantity_unit = price_unit[
        4:
    ]  # First three letters of a price unit give the currency (ISO 4217), fourth character is "/"
    if quantity_unit == market_quantity_unit:
        return 1
    elif quantity_unit == "MWh" and price_unit[4:] == "kWh":
        return 1000
    else:
        raise NotImplementedError


def determine_resolution(
    data: pd.DataFrame,
    forecasts: Optional[pd.DataFrame] = None,
    schedules: Optional[pd.DataFrame] = None,
) -> str:
    """
    Determine the resolution to be displayed under the plot.
    We try to get it from the DataFrame's meta data, or guess from the actual data.
    Lastly, we try the session.
    If nothing can be found this way, the resulting string is "?"
    """

    def determine_resolution_for(df: pd.DataFrame) -> str:
        if hasattr(df, "event_resolution"):  # BeliefsDataFrame
            freq_str = time_utils.timedelta_to_pandas_freq_str(df.event_resolution)
        elif hasattr(df.index, "freqstr") and df.index.freqstr is not None:
            freq_str = data.index.freqstr
        elif hasattr(df.index, "inferred_freq") and df.index.inferred_freq is not None:
            freq_str = df.index.inferred_freq
        elif "resolution" in session:
            freq_str = session["resolution"]
        else:
            return "?"
        return time_utils.freq_label_to_human_readable_label(freq_str)

    resolution = determine_resolution_for(data)
    if resolution == "?" and forecasts is not None:
        resolution = determine_resolution_for(forecasts)
    if resolution == "?" and schedules is not None:
        resolution = determine_resolution_for(schedules)
    return resolution



import pytest

from flask_security.utils import hash_password

from flexmeasures.data.services.users import create_user
from flexmeasures.data.models.assets import Asset
from flexmeasures.data.models.weather import WeatherSensor, WeatherSensorType
from flexmeasures.ui.tests.utils import login, logout


@pytest.fixture(scope="function")
def as_prosumer(client):
    """
    Login the default test prosumer and log him out afterwards.
    """
    login(client, "test_prosumer@seita.nl", "testtest")
    yield
    logout(client)


@pytest.fixture(scope="function")
def as_admin(client):
    """
    Login the admin user and log him out afterwards.
    """
    login(client, "flexmeasures-admin@seita.nl", "testtest")
    yield
    logout(client)


@pytest.fixture(scope="module", autouse=True)
def setup_ui_test_data(
    db,
    setup_account,
    setup_roles_users,
    setup_markets,
    setup_sources,
    setup_asset_types,
):
    """
    Create another prosumer, without data, and an admin
    Also, a weather sensor (and sensor type).
    """
    print("Setting up data for UI tests on %s" % db.engine)

    create_user(
        username="Site Admin",
        email="flexmeasures-admin@seita.nl",
        password=hash_password("testtest"),
        account_name=setup_account.name,
        user_roles=dict(name="admin", description="A site admin."),
    )

    test_prosumer2 = create_user(
        username="Second Test Prosumer",
        email="test_prosumer2@seita.nl",
        password=hash_password("testtest"),
        account_name=setup_account.name,
        user_roles=dict(
            name="Prosumer", description="A Prosumer with one asset but no data."
        ),
    )
    asset = Asset(
        name="solar pane 1",
        display_name="Solar Pane 1",
        asset_type_name="solar",
        unit="MW",
        capacity_in_mw=10,
        latitude=10,
        longitude=100,
        min_soc_in_mwh=0,
        max_soc_in_mwh=0,
        soc_in_mwh=0,
    )
    db.session.add(asset)
    asset.owner = test_prosumer2

    # Create 1 weather sensor
    test_sensor_type = WeatherSensorType(name="radiation")
    db.session.add(test_sensor_type)
    sensor = WeatherSensor(
        name="radiation_sensor",
        weather_sensor_type_name="radiation",
        latitude=33.4843866,
        longitude=126,
        unit="kW/m²",
    )
    db.session.add(sensor)

    print("Done setting up data for UI tests")



from typing import List, Tuple, Union

from flask import current_app
import pandas as pd
import numpy as np
from pandas.tseries.frequencies import to_offset
from pyomo.core import (
    ConcreteModel,
    Var,
    RangeSet,
    Param,
    Reals,
    Constraint,
    Objective,
    minimize,
)
from pyomo.environ import UnknownSolver  # noqa F401
from pyomo.environ import value
from pyomo.opt import SolverFactory

from flexmeasures.data.models.planning.utils import initialize_series

infinity = float("inf")


def device_scheduler(  # noqa C901
    device_constraints: List[pd.DataFrame],
    ems_constraints: pd.DataFrame,
    commitment_quantities: List[pd.Series],
    commitment_downwards_deviation_price: Union[List[pd.Series], List[float]],
    commitment_upwards_deviation_price: Union[List[pd.Series], List[float]],
) -> Tuple[List[pd.Series], float]:
    """Schedule devices given constraints on a device and EMS level, and given a list of commitments by the EMS.
    The commitments are assumed to be with regards to the flow of energy to the device (positive for consumption,
    negative for production). The solver minimises the costs of deviating from the commitments.

    Device constraints are on a device level. Handled constraints (listed by column name):
        max: maximum stock assuming an initial stock of zero (e.g. in MWh or boxes)
        min: minimum stock assuming an initial stock of zero
        equal: exact amount of stock (we do this by clamping min and max)
        derivative max: maximum flow (e.g. in MW or boxes/h)
        derivative min: minimum flow
        derivative equals: exact amount of flow (we do this by clamping derivative min and derivative max)
    EMS constraints are on an EMS level. Handled constraints (listed by column name):
        derivative max: maximum flow
        derivative min: minimum flow
    Commitments are on an EMS level. Parameter explanations:
        commitment_quantities: amounts of flow specified in commitments (both previously ordered and newly requested)
            - e.g. in MW or boxes/h
        commitment_downwards_deviation_price: penalty for downwards deviations of the flow
            - e.g. in EUR/MW or EUR/(boxes/h)
            - either a single value (same value for each flow value) or a Series (different value for each flow value)
        commitment_upwards_deviation_price: penalty for upwards deviations of the flow

    All Series and DataFrames should have the same resolution.

    For now we pass in the various constraints and prices as separate variables, from which we make a MultiIndex
    DataFrame. Later we could pass in a MultiIndex DataFrame directly.
    """

    # If the EMS has no devices, don't bother
    if len(device_constraints) == 0:
        return [], 0

    # Check if commitments have the same time window and resolution as the constraints
    start = device_constraints[0].index.to_pydatetime()[0]
    resolution = pd.to_timedelta(device_constraints[0].index.freq)
    end = device_constraints[0].index.to_pydatetime()[-1] + resolution
    if len(commitment_quantities) != 0:
        start_c = commitment_quantities[0].index.to_pydatetime()[0]
        resolution_c = pd.to_timedelta(commitment_quantities[0].index.freq)
        end_c = commitment_quantities[0].index.to_pydatetime()[-1] + resolution
        if not (start_c == start and end_c == end):
            raise Exception(
                "Not implemented for different time windows.\n(%s,%s)\n(%s,%s)"
                % (start, end, start_c, end_c)
            )
        if resolution_c != resolution:
            raise Exception(
                "Not implemented for different resolutions.\n%s\n%s"
                % (resolution, resolution_c)
            )

    # Turn prices per commitment into prices per commitment flow
    if len(commitment_downwards_deviation_price) != 0:
        if all(
            isinstance(price, float) for price in commitment_downwards_deviation_price
        ):
            commitment_downwards_deviation_price = [
                initialize_series(price, start, end, resolution)
                for price in commitment_downwards_deviation_price
            ]
    if len(commitment_upwards_deviation_price) != 0:
        if all(
            isinstance(price, float) for price in commitment_upwards_deviation_price
        ):
            commitment_upwards_deviation_price = [
                initialize_series(price, start, end, resolution)
                for price in commitment_upwards_deviation_price
            ]

    model = ConcreteModel()

    # Add indices for devices (d), datetimes (j) and commitments (c)
    model.d = RangeSet(0, len(device_constraints) - 1, doc="Set of devices")
    model.j = RangeSet(
        0, len(device_constraints[0].index.to_pydatetime()) - 1, doc="Set of datetimes"
    )
    model.c = RangeSet(0, len(commitment_quantities) - 1, doc="Set of commitments")

    # Add parameters
    def price_down_select(m, c, j):
        return commitment_downwards_deviation_price[c].iloc[j]

    def price_up_select(m, c, j):
        return commitment_upwards_deviation_price[c].iloc[j]

    def commitment_quantity_select(m, c, j):
        return commitment_quantities[c].iloc[j]

    def device_max_select(m, d, j):
        max_v = device_constraints[d]["max"].iloc[j]
        equal_v = device_constraints[d]["equals"].iloc[j]
        if np.isnan(max_v) and np.isnan(equal_v):
            return infinity
        else:
            return np.nanmin([max_v, equal_v])

    def device_min_select(m, d, j):
        min_v = device_constraints[d]["min"].iloc[j]
        equal_v = device_constraints[d]["equals"].iloc[j]
        if np.isnan(min_v) and np.isnan(equal_v):
            return -infinity
        else:
            return np.nanmax([min_v, equal_v])

    def device_derivative_max_select(m, d, j):
        max_v = device_constraints[d]["derivative max"].iloc[j]
        equal_v = device_constraints[d]["derivative equals"].iloc[j]
        if np.isnan(max_v) and np.isnan(equal_v):
            return infinity
        else:
            return np.nanmin([max_v, equal_v])

    def device_derivative_min_select(m, d, j):
        min_v = device_constraints[d]["derivative min"].iloc[j]
        equal_v = device_constraints[d]["derivative equals"].iloc[j]
        if np.isnan(min_v) and np.isnan(equal_v):
            return -infinity
        else:
            return np.nanmax([min_v, equal_v])

    def ems_derivative_max_select(m, j):
        v = ems_constraints["derivative max"].iloc[j]
        if np.isnan(v):
            return infinity
        else:
            return v

    def ems_derivative_min_select(m, j):
        v = ems_constraints["derivative min"].iloc[j]
        if np.isnan(v):
            return -infinity
        else:
            return v

    model.up_price = Param(model.c, model.j, initialize=price_up_select)
    model.down_price = Param(model.c, model.j, initialize=price_down_select)
    model.commitment_quantity = Param(
        model.c, model.j, initialize=commitment_quantity_select
    )
    model.device_max = Param(model.d, model.j, initialize=device_max_select)
    model.device_min = Param(model.d, model.j, initialize=device_min_select)
    model.device_derivative_max = Param(
        model.d, model.j, initialize=device_derivative_max_select
    )
    model.device_derivative_min = Param(
        model.d, model.j, initialize=device_derivative_min_select
    )
    model.ems_derivative_max = Param(model.j, initialize=ems_derivative_max_select)
    model.ems_derivative_min = Param(model.j, initialize=ems_derivative_min_select)

    # Add variables
    model.power = Var(model.d, model.j, domain=Reals, initialize=0)

    # Add constraints as a tuple of (lower bound, value, upper bound)
    def device_bounds(m, d, j):
        return (
            m.device_min[d, j],
            sum(m.power[d, k] for k in range(0, j + 1)),
            m.device_max[d, j],
        )

    def device_derivative_bounds(m, d, j):
        return (
            m.device_derivative_min[d, j],
            m.power[d, j],
            m.device_derivative_max[d, j],
        )

    def ems_derivative_bounds(m, j):
        return m.ems_derivative_min[j], sum(m.power[:, j]), m.ems_derivative_max[j]

    model.device_energy_bounds = Constraint(model.d, model.j, rule=device_bounds)
    model.device_power_bounds = Constraint(
        model.d, model.j, rule=device_derivative_bounds
    )
    model.ems_power_bounds = Constraint(model.j, rule=ems_derivative_bounds)

    # Add objective
    def cost_function(m):
        costs = 0
        for c in m.c:
            for j in m.j:
                ems_power_in_j = sum(m.power[d, j] for d in m.d)
                ems_power_deviation = ems_power_in_j - m.commitment_quantity[c, j]
                if value(ems_power_deviation) >= 0:
                    costs += ems_power_deviation * m.up_price[c, j]
                else:
                    costs += ems_power_deviation * m.down_price[c, j]
        return costs

    model.costs = Objective(rule=cost_function, sense=minimize)

    # Solve
    SolverFactory(current_app.config.get("FLEXMEASURES_LP_SOLVER")).solve(model)

    planned_costs = value(model.costs)
    planned_power_per_device = []
    for d in model.d:
        planned_device_power = [model.power[d, j].value for j in model.j]
        planned_power_per_device.append(
            pd.Series(
                index=pd.date_range(
                    start=start, end=end, freq=to_offset(resolution), closed="left"
                ),
                data=planned_device_power,
            )
        )

    # model.pprint()
    # print(planned_costs)
    # input()
    return planned_power_per_device, planned_costs



from typing import Optional

from marshmallow import validates, ValidationError, fields

from flexmeasures.data import ma
from flexmeasures.data.models.user import Account
from flexmeasures.data.models.generic_assets import GenericAsset, GenericAssetType


class GenericAssetSchema(ma.SQLAlchemySchema):
    """
    GenericAsset schema, with validations.
    """

    id = ma.auto_field()
    name = fields.Str()
    account_id = ma.auto_field()
    latitude = ma.auto_field()
    longitude = ma.auto_field()
    generic_asset_type_id = fields.Integer()

    class Meta:
        model = GenericAsset

    @validates("generic_asset_type_id")
    def validate_generic_asset_type(self, generic_asset_type_id: int):
        generic_asset_type = GenericAssetType.query.get(generic_asset_type_id)
        if not generic_asset_type:
            raise ValidationError(
                f"GenericAssetType with id {generic_asset_type_id} doesn't exist."
            )

    @validates("account_id")
    def validate_account(self, account_id: int):
        account = Account.query.get(account_id)
        if not account:
            raise ValidationError(f"Account with Id {account_id} doesn't exist.")

    @validates("latitude")
    def validate_latitude(self, latitude: Optional[float]):
        """Validate optional latitude."""
        if latitude is None:
            return
        if latitude < -90:
            raise ValidationError(
                f"Latitude {latitude} exceeds the minimum latitude of -90 degrees."
            )
        if latitude > 90:
            raise ValidationError(
                f"Latitude {latitude} exceeds the maximum latitude of 90 degrees."
            )

    @validates("longitude")
    def validate_longitude(self, longitude: Optional[float]):
        """Validate optional longitude."""
        if longitude is None:
            return
        if longitude < -180:
            raise ValidationError(
                f"Longitude {longitude} exceeds the minimum longitude of -180 degrees."
            )
        if longitude > 180:
            raise ValidationError(
                f"Longitude {longitude} exceeds the maximum longitude of 180 degrees."
            )


class GenericAssetTypeSchema(ma.SQLAlchemySchema):
    """
    GenericAssetType schema, with validations.
    """

    id = ma.auto_field()
    name = fields.Str()
    description = ma.auto_field()

    class Meta:
        model = GenericAssetType



import os

from flask import Flask
from flask_migrate import Migrate
from flask_marshmallow import Marshmallow

from flexmeasures.data.config import configure_db_for, db
from flexmeasures.data.auth_setup import configure_auth
from flexmeasures.data.transactional import after_request_exception_rollback_session


ma: Marshmallow = Marshmallow()


def register_at(app: Flask):
    # First configure the central db object and Alembic's migration tool
    configure_db_for(app)
    Migrate(app, db, directory=os.path.join(app.root_path, "data", "migrations"))

    global ma
    ma.init_app(app)

    configure_auth(app, db)

    if app.cli:
        # Register some useful custom scripts with the flask cli
        with app.app_context():
            import flexmeasures.data.scripts.cli_tasks.jobs
            import flexmeasures.data.scripts.cli_tasks.monitor
            import flexmeasures.data.scripts.cli_tasks.data_add
            import flexmeasures.data.scripts.cli_tasks.data_delete
            import flexmeasures.data.scripts.cli_tasks.db_ops
            import flexmeasures.data.scripts.cli_tasks.testing  # noqa: F401

    app.teardown_request(after_request_exception_rollback_session)



"""Tests a small simulation against FlexMeasures running on a server."""
import sys
from datetime import timedelta

from isodate import parse_datetime

from flexmeasures.data.scripts.simulation_utils import (
    check_version,
    check_services,
    get_auth_token,
    get_connections,
    post_meter_data,
    post_price_forecasts,
    post_weather_data,
)


# Setup host - pass hostname to overwrite
host = "http://localhost:5000"
if len(sys.argv) > 1:
    host = sys.argv[1]

latest_version = check_version(host)
services = check_services(host, latest_version)
auth_token = get_auth_token(host)
connections = get_connections(host, latest_version, auth_token)

# Initialisation
num_days = 50
sim_start = "2018-01-01T00:00:00+09:00"
post_price_forecasts(
    host, latest_version, auth_token, start=parse_datetime(sim_start), num_days=num_days
)
post_weather_data(
    host, latest_version, auth_token, start=parse_datetime(sim_start), num_days=num_days
)
post_meter_data(
    host,
    latest_version,
    auth_token,
    start=parse_datetime(sim_start),
    num_days=num_days,
    connection=connections[1],
)

# Run forecasting jobs
input(
    "Run all forecasting jobs, then press Enter to continue ...\n"
    "You can run this in another flexmeasures-venv shell:\n\n"
    "flask run_forecasting_worker\n"
)

# Steps
num_steps = 10
for i in range(num_steps):
    print("Running step %s out of %s" % (i, num_steps))

    post_weather_data(
        host,
        latest_version,
        auth_token,
        start=parse_datetime(sim_start) + timedelta(days=num_days + i),
        num_days=1,
    )
    post_meter_data(
        host,
        latest_version,
        auth_token,
        start=parse_datetime(sim_start) + timedelta(days=num_days + i),
        num_days=1,
        connection=connections[1],
    )

    # Run forecasting jobs
    input("Run all forecasting jobs again and press enter to continue..\n")



from flexmeasures.app import create as create_app
from flexmeasures.utils import config_defaults

"""
Run the FlexMeasures application locally.

Best to use in a development setup. A professional web server should be handed the app object to use in a WSGI context.
"""

if __name__ == "__main__":

    print("Starting the FlexMeasures Platform ...")

    create_app().run(debug=config_defaults.DevelopmentConfig.DEBUG, load_dotenv=False)
